from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import ImprovementEnvBase, RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_distance, get_tour_length
from rl4co.utils.pylogger import get_pylogger

from .generator import TSPGenerator

try:
    from .local_search import local_search
except ImportError:
    # In case some dependencies are not installed (e.g., numba)
    local_search = None
from .render import render, render_improvement

log = get_pylogger(__name__)


class TSPEnv(RL4COEnvBase):
    """Traveling Salesman Problem (TSP) environment
    At each step, the agent chooses a city to visit. The reward is 0 unless the agent visits all the cities.
    In that case, the reward is (-)length of the path: maximizing the reward is equivalent to minimizing the path length.

    Observations:
        - locations of each customer.
        - the current location of the vehicle.

    Constraints:
        - the tour must return to the starting customer.
        - each customer must be visited exactly once.

    Finish condition:
        - the agent has visited all customers and returned to the starting customer.

    Reward:
        - (minus) the negative length of the path.

    Args:
        generator: TSPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "tsp"

    def __init__(
        self,
        generator: TSPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = TSPGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    @staticmethod
    def _step(td: TensorDict) -> TensorDict:
        current_node = td["action"]
        first_node = current_node if td["i"].all() == 0 else td["first_node"]

        # # Set not visited to 0 (i.e., we visited the node)
        available = td["action_mask"].scatter(
            -1, current_node.unsqueeze(-1).expand_as(td["action_mask"]), 0
        )

        # We are done there are no unvisited locations
        done = torch.sum(available, dim=-1) == 0

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)

        td.update(
            {
                "first_node": first_node,
                "current_node": current_node,
                "i": td["i"] + 1,
                "action_mask": available,
                "reward": reward,
                "done": done,
            },
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        # Initialize locations
        device = td.device
        init_locs = td["locs"]

        # We do not enforce loading from self for flexibility
        num_loc = init_locs.shape[-2]

        # Other variables
        current_node = torch.zeros((batch_size), dtype=torch.int64, device=device)
        available = torch.ones(
            (*batch_size, num_loc), dtype=torch.bool, device=device
        )  # 1 means not visited, i.e. action is allowed
        i = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)

        return TensorDict(
            {
                "locs": init_locs,
                "first_node": current_node,
                "current_node": current_node,
                "i": i,
                "action_mask": available,
                "reward": torch.zeros((*batch_size, 1), dtype=torch.float32),
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: TSPGenerator):
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc, 2),
                dtype=torch.float32,
            ),
            first_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc,
        )
        self.reward_spec = Unbounded(shape=(1))
        self.done_spec = Unbounded(shape=(1), dtype=torch.bool)

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        if self.check_solution:
            self.check_solution_validity(td, actions)

        # Gather locations in order of tour and return distance between them (i.e., -reward)
        locs_ordered = gather_by_index(td["locs"], actions)
        return -get_tour_length(locs_ordered)

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor) -> None:
        """Check that solution is valid: nodes are visited exactly once"""
        assert (
            torch.arange(actions.size(1), out=actions.data.new())
            .view(1, -1)
            .expand_as(actions)
            == actions.data.sort(1)[0]
        ).all(), "Invalid tour"

    def replace_selected_actions(
        self,
        cur_actions: torch.Tensor,
        new_actions: torch.Tensor,
        selection_mask: torch.Tensor,
    ) -> torch.Tensor:
        """
        Replace selected current actions with updated actions based on `selection_mask`.

        Args:
            cur_actions [batch_size, num_loc]
            new_actions [batch_size, num_loc]
            selection_mask [batch_size,]
        """
        cur_actions[selection_mask] = new_actions[selection_mask]
        return cur_actions

    @staticmethod
    def local_search(td: TensorDict, actions: torch.Tensor, **kwargs) -> torch.Tensor:
        assert (
            local_search is not None
        ), "Cannot import local_search module. Check if `numba` is installed."
        return local_search(td, actions, **kwargs)

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)


class TSPkoptEnv(ImprovementEnvBase):
    """Traveling Salesman Problem (PDP) environment for performing the neural k-opt search.

    The goal is to search for optimal solutions to TSP by performing a k-opt neighborhood search on a given initial solution.

    Observations:
        - locations of each customer
        - current solution to be improved
        - the current step

    Constraints:
        - the tour must return to the starting customer.
        - each customer must be visited exactly once.

    Finish condition:
        - None

    Reward:
        - the immediate reduced cost over the current best-so-far solution

    Args:
        generator: TSPGenerator instance as the data generator
        generator_params: parameters for the generator
        k_max: the maximum k value for k-opt:
            if k_max==2, the MDP in DACT(https://arxiv.org/abs/2110.02544) is used;
            if k_max>2, the MDP in NeuOpt(https://arxiv.org/abs/2310.18264) is used;
    """

    name = "tsp_kopt"

    def __init__(
        self,
        generator: TSPGenerator = None,
        generator_params: dict = {},
        k_max: int = 2,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = TSPGenerator(**generator_params)
        self.generator = generator
        self.k_max = k_max
        self.two_opt_mode = k_max == 2
        self._make_spec(self.generator)

    def _step(self, td: TensorDict, solution_to=None) -> TensorDict:
        # get state information from td
        solution_best = td["rec_best"]
        locs = td["locs"]
        cost_bsf = td["cost_bsf"]
        bs, gs = solution_best.size()

        # perform local_operator
        if solution_to is None:
            action = td["action"]
            solution = td["rec_current"]
            next_rec = self._local_operator(solution, action)
        else:
            next_rec = solution_to.clone()
        new_obj = self.get_costs(locs, next_rec)

        # compute reward and update best-so-far solutions
        now_bsf = torch.where(new_obj < cost_bsf, new_obj, cost_bsf)
        reward = cost_bsf - now_bsf
        index = reward > 0.0
        solution_best[index] = next_rec[index].clone()

        # reset visited_time
        visited_time = td["visited_time"] * 0
        pre = torch.zeros((bs), device=visited_time.device).long()
        arange = torch.arange(bs)
        for i in range(gs):
            current_nodes = next_rec[arange, pre]
            visited_time[arange, current_nodes] = i + 1
            pre = current_nodes
        visited_time = visited_time.long()

        # Update step
        td.update(
            {
                "cost_current": new_obj,
                "cost_bsf": now_bsf,
                "rec_current": next_rec,
                "rec_best": solution_best,
                "visited_time": visited_time,
                "i": td["i"] + 1 if solution_to is None else td["i"],
                "reward": reward,
            }
        )

        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        device = td.device

        locs = td["locs"]
        current_rec = self.generator._get_initial_solutions(locs).to(device)

        obj = self.get_costs(locs, current_rec)

        # get index according to the solutions in the linked list data structure
        bs = batch_size[0]
        seq_length = self.generator.num_loc
        visited_time = torch.zeros((bs, seq_length)).to(device)
        pre = torch.zeros((bs)).to(device).long()
        arange = torch.arange(bs)
        for i in range(seq_length):
            current_nodes = current_rec[arange, pre]
            visited_time[arange, current_nodes] = i + 1
            pre = current_nodes
        visited_time = visited_time.long()

        i = torch.zeros((*batch_size, 1), dtype=torch.int64).to(device)

        return TensorDict(
            {
                "locs": locs,
                "cost_current": obj,
                "cost_bsf": obj.clone(),
                "rec_current": current_rec,
                "rec_best": current_rec.clone(),
                "visited_time": visited_time,
                "i": i,
            },
            batch_size=batch_size,
        )

    def _local_operator(self, solution, action):
        rec = solution.clone()

        if self.two_opt_mode:
            # get actions
            first = action[:, 0].view(-1, 1)
            second = action[:, 1].view(-1, 1)

            # fix connection for first node
            argsort = solution.argsort()
            pre_first = argsort.gather(1, first)
            pre_first = torch.where(pre_first != second, pre_first, first)
            rec.scatter_(1, pre_first, second)

            # fix connection for second node
            post_second = solution.gather(1, second)
            post_second = torch.where(post_second != first, post_second, second)
            rec.scatter_(1, first, post_second)

            # reverse loop:
            cur = first
            for i in range(self.generator.num_loc):
                cur_next = solution.gather(1, cur)
                rec.scatter_(
                    1, cur_next, torch.where(cur != second, cur, rec.gather(1, cur_next))
                )
                cur = torch.where(cur != second, cur_next, cur)

            rec_next = rec

        else:
            # action bs * (K_index, K_from, K_to)
            selected_index = action[:, : self.k_max]
            left = action[:, self.k_max : 2 * self.k_max]
            right = action[:, 2 * self.k_max :]

            # prepare
            rec_next = rec.clone()
            right_nodes = rec.gather(1, selected_index)
            argsort = rec.argsort()

            # new rec
            rec_next.scatter_(1, left, right)
            cur = left[:, :1].clone()
            for i in range(
                self.generator.num_loc - 2
            ):  # self.generator.num_loc - 2 is already correct
                next_cur = rec_next.gather(1, cur)
                pre_next_wrt_old = argsort.gather(1, next_cur)
                reverse_link_condition = (cur != pre_next_wrt_old) & ~(
                    (next_cur == right_nodes).any(-1, True)
                )
                next_next_cur = rec_next.gather(1, next_cur)
                rec_next.scatter_(
                    1,
                    next_cur,
                    torch.where(reverse_link_condition, pre_next_wrt_old, next_next_cur),
                )
                # if i >= self.generator.num_loc - 2: assert (reverse_link_condition == False).all()
                cur = next_cur

        return rec_next

    def _make_spec(self, generator: TSPGenerator):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc, 2),
                dtype=torch.float32,
            ),
            cost_current=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            cost_bsf=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            rec_current=Unbounded(
                shape=(self.generator.num_loc),
                dtype=torch.int64,
            ),
            rec_best=Unbounded(
                shape=(self.generator.num_loc),
                dtype=torch.int64,
            ),
            visited_time=Unbounded(
                shape=(self.generator.num_loc, self.generator.num_loc),
                dtype=torch.int64,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(self.k_max * 3,),
            dtype=torch.int64,
            low=0,
            high=self.generator.num_loc,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def check_solution_validity(self, td, actions=None):
        # The function can be called by the agent to check the validity of the best found solution
        # Note that the args actions are not used in improvement method.

        solution = td["rec_best"]
        batch_size, graph_size = solution.size()

        assert (
            torch.arange(graph_size, out=solution.data.new())
            .view(1, -1)
            .expand_as(solution)
            == solution.data.sort(1)[0]
        ).all(), "Not visiting all nodes"

    def get_mask(self, td):
        # return mask that is 1 if the corresponding action is feasible, 0 otherwise
        visited_time = td["visited_time"]
        bs, gs = visited_time.size()
        if self.two_opt_mode:
            selfmask = torch.eye(gs).view(1, gs, gs).to(td.device)
            masks = selfmask.expand(bs, gs, gs).bool()
            return ~masks
        else:
            assert False, "The masks for NeuOpt are handled within its policy"

    def _random_action(self, td):
        bs, gs = td["rec_best"].size()

        if self.two_opt_mode:
            mask = self.get_mask(td)
            logits = torch.rand(bs, gs, gs)
            logits[~mask] = -1e20
            prob = torch.softmax(logits.view(bs, -1), -1)
            sample = prob.multinomial(1)
            td["action"] = torch.cat((sample // (gs), sample % (gs)), -1)

        else:
            rec = td["rec_current"]
            visited_time = td["visited_time"]
            action_index = torch.zeros(bs, self.k_max, dtype=torch.long)
            k_action_left = torch.zeros(bs, self.k_max + 1, dtype=torch.long)
            k_action_right = torch.zeros(bs, self.k_max, dtype=torch.long)
            next_of_last_action = torch.zeros((bs, 1), dtype=torch.long) - 1
            mask = torch.zeros((bs, gs), dtype=torch.bool)
            stopped = torch.ones(bs, dtype=torch.bool)

            for i in range(self.k_max):
                # Sample action for a_i
                logits = torch.rand(bs, gs)
                logits[mask.clone()] = -1e30
                prob = torch.softmax(logits, -1)
                action = prob.multinomial(1)
                value_max, action_max = prob.max(-1, True)  ### fix bug of pytorch
                action = torch.where(
                    1 - value_max.view(-1, 1) < 1e-5, action_max.view(-1, 1), action
                )  ### fix bug of pytorch
                if i > 0:
                    action = torch.where(
                        stopped.unsqueeze(-1), action_index[:, :1], action
                    )

                # Store and Process actions
                next_of_new_action = rec.gather(1, action)
                action_index[:, i] = action.squeeze().clone()
                k_action_left[stopped, i] = action[stopped].squeeze().clone()
                k_action_right[~stopped, i - 1] = action[~stopped].squeeze().clone()
                k_action_left[:, i + 1] = next_of_new_action.squeeze().clone()

                # Process if k-opt close
                if i > 0:
                    stopped = stopped | (action == next_of_last_action).squeeze()
                else:
                    stopped = (action == next_of_last_action).squeeze()
                k_action_left[stopped, i] = k_action_left[stopped, i - 1]
                k_action_right[stopped, i] = k_action_right[stopped, i - 1]

                # Calc next basic masks
                if i == 0:
                    visited_time_tag = (
                        visited_time - visited_time.gather(1, action)
                    ) % gs
                mask &= False
                mask[(visited_time_tag <= visited_time_tag.gather(1, action))] = True
                if i == 0:
                    mask[visited_time_tag > (gs - 2)] = True
                mask[stopped, action[stopped].squeeze()] = (
                    False  # allow next k-opt starts immediately
                )
                # if True:#i == self.k_max - 2: # allow special case: close k-opt at the first selected node
                index_allow_first_node = (~stopped) & (
                    next_of_new_action.squeeze() == action_index[:, 0]
                )
                mask[index_allow_first_node, action_index[index_allow_first_node, 0]] = (
                    False
                )

                # Move to next
                next_of_last_action = next_of_new_action
                next_of_last_action[stopped] = -1

            # Form final action
            k_action_right[~stopped, -1] = k_action_left[~stopped, -1].clone()
            k_action_left = k_action_left[:, : self.k_max]
            td["action"] = torch.cat((action_index, k_action_left, k_action_right), -1)

        return td["action"]

    @classmethod
    def render(cls, td: TensorDict, actions: torch.Tensor = None, ax=None):
        solution_current = cls.get_current_solution(td)
        solution_best = cls.get_best_solution(td)
        return render_improvement(td, solution_current, solution_best)


class DenseRewardTSPEnv(TSPEnv):
    """
    This is an experimental version of the TSPEnv to be used with stepwise PPO. That is
    this environment defines a stepwise reward function for the TSP which is the distance added
    to the current tour by the given action.
    """

    def __init__(
        self, generator: TSPGenerator = None, generator_params: dict = {}, **kwargs
    ):
        super().__init__(
            generator,
            generator_params,
            check_solution=False,
            _torchrl_mode=True,
            **kwargs,
        )

    def _step(self, td):
        last_node = td["current_node"].clone()
        current_node = td["action"]

        first_node = current_node if td["i"].all() == 0 else td["first_node"]

        # # Set not visited to 0 (i.e., we visited the node)
        available = td["action_mask"].scatter(
            -1, current_node.unsqueeze(-1).expand_as(td["action_mask"]), 0
        )

        # We are done there are no unvisited locations
        done = torch.sum(available, dim=-1) == 0

        # calc stepwise reward
        last_node_loc = gather_by_index(td["locs"], last_node)
        curr_node_loc = gather_by_index(td["locs"], current_node)
        reward = get_distance(last_node_loc, curr_node_loc)[:, None]

        td.update(
            {
                "first_node": first_node,
                "current_node": current_node,
                "i": td["i"] + 1,
                "action_mask": available,
                "reward": reward,
                "done": done,
            },
        )
        return td

    def _get_reward(self, td, actions=None) -> TensorDict:
        if actions is not None:
            # Gather locations in order of tour and return distance between them (i.e., -reward)
            locs_ordered = gather_by_index(td["locs"], actions)
            return -get_tour_length(locs_ordered)
        return -td["reward"]



from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


# From Kool et al. 2019, Hottung et al. 2022, Kim et al. 2023
CAPACITIES = {
    10: 20.0,
    15: 25.0,
    20: 30.0,
    30: 33.0,
    40: 37.0,
    50: 40.0,
    60: 43.0,
    75: 45.0,
    100: 50.0,
    125: 55.0,
    150: 60.0,
    200: 70.0,
    500: 100.0,
    1000: 150.0,
}


class CVRPGenerator(Generator):
    """Data generator for the Capacitated Vehicle Routing Problem (CVRP).

    Args:
        num_loc: number of locations (cities) in the VRP, without the depot. (e.g. 10 means 10 locs + 1 depot)
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates
        loc_distribution: distribution for the location coordinates
        depot_distribution: distribution for the depot location. If None, sample the depot from the locations
        min_demand: minimum value for the demand of each customer
        max_demand: maximum value for the demand of each customer
        demand_distribution: distribution for the demand of each customer
        capacity: capacity of the vehicle

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer
            depot [batch_size, 2]: location of the depot
            demand [batch_size, num_loc]: demand of each customer
            capacity [batch_size]: capacity of the vehicle
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: int | float | str | type | Callable = Uniform,
        depot_distribution: int | float | str | type | Callable = None,
        min_demand: int = 1,
        max_demand: int = 10,
        demand_distribution: int | float | type | Callable = Uniform,
        vehicle_capacity: float = 1.0,
        capacity: float = None,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.min_demand = min_demand
        self.max_demand = max_demand
        self.vehicle_capacity = vehicle_capacity

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

        # Depot distribution
        if kwargs.get("depot_sampler", None) is not None:
            self.depot_sampler = kwargs["depot_sampler"]
        else:
            self.depot_sampler = (
                get_sampler("depot", depot_distribution, min_loc, max_loc, **kwargs)
                if depot_distribution is not None
                else None
            )

        # Demand distribution
        if kwargs.get("demand_sampler", None) is not None:
            self.demand_sampler = kwargs["demand_sampler"]
        else:
            self.demand_sampler = get_sampler(
                "demand", demand_distribution, min_demand - 1, max_demand - 1, **kwargs
            )

        # Capacity
        if (
            capacity is None
        ):  # If not provided, use the default capacity from Kool et al. 2019
            capacity = CAPACITIES.get(num_loc, None)
        if (
            capacity is None
        ):  # If not in the table keys, find the closest number of nodes as the key
            closest_num_loc = min(CAPACITIES.keys(), key=lambda x: abs(x - num_loc))
            capacity = CAPACITIES[closest_num_loc]
            log.warning(
                f"The capacity capacity for {num_loc} locations is not defined. Using the closest capacity: {capacity}\
                    with {closest_num_loc} locations."
            )
        self.capacity = capacity

    def _generate(self, batch_size) -> TensorDict:

        # Sample locations: depot and customers
        if self.depot_sampler is not None:
            depot = self.depot_sampler.sample((*batch_size, 2))
            locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))
        else:
            # if depot_sampler is None, sample the depot from the locations
            locs = self.loc_sampler.sample((*batch_size, self.num_loc + 1, 2))
            depot = locs[..., 0, :]
            locs = locs[..., 1:, :]

        # Sample demands
        demand = self.demand_sampler.sample((*batch_size, self.num_loc))
        demand = (demand.int() + 1).float()

        # Sample capacities
        capacity = torch.full((*batch_size, 1), self.capacity)

        return TensorDict(
            {
                "locs": locs,
                "depot": depot,
                "demand": demand / self.capacity,
                "capacity": capacity,
            },
            batch_size=batch_size,
        )



from dataclasses import dataclass
from typing import Tuple

import torch
import torch.nn as nn

from tensordict import TensorDict
from torch import Tensor

from rl4co.models.nn.env_embeddings.context import FFSPContext
from rl4co.models.zoo.am.decoder import AttentionModelDecoder
from rl4co.utils.decoding import decode_logprobs, process_logits
from rl4co.utils.ops import gather_by_index


@dataclass
class PrecomputedCache:
    node_embeddings: Tensor | TensorDict
    graph_context: Tensor | float
    glimpse_key: Tensor
    glimpse_val: Tensor
    logit_key: Tensor


class MatNetDecoder(AttentionModelDecoder):
    def _precompute_cache(self, embeddings: Tuple[Tensor, Tensor], *args, **kwargs):
        row_emb, col_emb = embeddings
        (
            glimpse_key_fixed,
            glimpse_val_fixed,
            logit_key,
        ) = self.project_node_embeddings(
            col_emb
        ).chunk(3, dim=-1)

        # Optionally disable the graph context from the initial embedding as done in POMO
        if self.use_graph_context:
            graph_context = self.project_fixed_context(col_emb.mean(1))
        else:
            graph_context = 0

        # Organize in a dataclass for easy access
        return PrecomputedCache(
            node_embeddings=row_emb,
            graph_context=graph_context,
            glimpse_key=glimpse_key_fixed,
            glimpse_val=glimpse_val_fixed,
            logit_key=logit_key,
        )


class MatNetFFSPDecoder(AttentionModelDecoder):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        linear_bias: bool = False,
        out_bias_pointer_attn: bool = True,
        use_graph_context: bool = False,
        **kwargs,
    ):
        context_embedding = FFSPContext(embed_dim)

        super().__init__(
            env_name="ffsp",
            embed_dim=embed_dim,
            num_heads=num_heads,
            context_embedding=context_embedding,
            out_bias_pointer_attn=out_bias_pointer_attn,
            linear_bias=linear_bias,
            use_graph_context=use_graph_context,
            **kwargs,
        )

        self.no_job_emb = nn.Parameter(torch.rand(1, 1, embed_dim), requires_grad=True)

    def _precompute_cache(self, embeddings: Tuple[Tensor, Tensor], **kwargs):
        job_emb, ma_emb = embeddings

        bs, _, emb_dim = job_emb.shape

        job_emb_plus_one = torch.cat(
            (job_emb, self.no_job_emb.expand((bs, 1, emb_dim))), dim=1
        )

        (
            glimpse_key_fixed,
            glimpse_val_fixed,
            logit_key,
        ) = self.project_node_embeddings(
            job_emb_plus_one
        ).chunk(3, dim=-1)

        # Optionally disable the graph context from the initial embedding as done in POMO
        if self.use_graph_context:
            graph_context = self.project_fixed_context(job_emb_plus_one.mean(1))
        else:
            graph_context = 0

        embeddings = TensorDict(
            {"job_embeddings": job_emb_plus_one, "machine_embeddings": ma_emb},
            batch_size=bs,
        )
        # Organize in a dataclass for easy access
        return PrecomputedCache(
            node_embeddings=embeddings,
            graph_context=graph_context,
            glimpse_key=glimpse_key_fixed,
            glimpse_val=glimpse_val_fixed,
            logit_key=logit_key,
        )


class MultiStageFFSPDecoder(MatNetFFSPDecoder):
    """Decoder class for the solving the FFSP using a seperate MatNet decoder for each stage
    as originally implemented by Kwon et al. (2021)
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        use_graph_context: bool = True,
        tanh_clipping: float = 10,
        **kwargs,
    ):
        super().__init__(
            embed_dim=embed_dim,
            num_heads=num_heads,
            use_graph_context=use_graph_context,
            **kwargs,
        )
        self.cached_embs: PrecomputedCache = None
        self.tanh_clipping = tanh_clipping

    def _precompute_cache(self, embeddings: Tuple[Tensor], **kwargs):
        self.cached_embs = super()._precompute_cache(embeddings, **kwargs)

    def forward(
        self,
        td: TensorDict,
        decode_type="sampling",
        num_starts: int = 1,
        **decoding_kwargs,
    ) -> Tuple[Tensor, Tensor, TensorDict]:

        logits, mask = super().forward(td, self.cached_embs, num_starts)
        logprobs = process_logits(
            logits,
            mask,
            tanh_clipping=self.tanh_clipping,
            **decoding_kwargs,
        )
        job_selected = decode_logprobs(logprobs, mask, decode_type)
        job_prob = gather_by_index(logprobs, job_selected, dim=1)

        return job_selected, job_prob



import os

from functools import partial

import numpy as np
import torch

from tensordict.tensordict import TensorDict
from torch.nn.functional import one_hot

from rl4co.envs.common.utils import Generator
from rl4co.utils.pylogger import get_pylogger

from .parser import get_max_ops_from_files, read

log = get_pylogger(__name__)


class JSSPGenerator(Generator):
    """Data generator for the Job-Shop Scheduling Problem (JSSP)

    Args:
        num_stage: number of stages
        num_machine: number of machines
        num_job: number of jobs
        min_time: minimum running time of each job on each machine
        max_time: maximum running time of each job on each machine
        flatten_stages: whether to flatten the stages
        one2one_ma_map: whether each machine should have exactly one operation per job (common in jssp benchmark instances)

    Returns:
        A TensorDict with the following key:
            start_op_per_job [batch_size, num_jobs]: first operation of each job
            end_op_per_job [batch_size, num_jobs]: last operation of each job
            proc_times [batch_size, num_machines, total_n_ops]: processing time of ops on machines
            pad_mask [batch_size, total_n_ops]: not all instances have the same number of ops, so padding is used

    """

    def __init__(
        self,
        num_jobs: int = 6,
        num_machines: int = 6,
        min_ops_per_job: int = None,
        max_ops_per_job: int = None,
        min_processing_time: int = 1,
        max_processing_time: int = 99,
        one2one_ma_map: bool = True,
        **unused_kwargs,
    ):
        self.num_jobs = num_jobs
        self.num_mas = num_machines
        # quite common in jssp to have as many ops per job as there are machines
        self.min_ops_per_job = min_ops_per_job or self.num_mas
        self.max_ops_per_job = max_ops_per_job or self.num_mas
        self.min_processing_time = min_processing_time
        self.max_processing_time = max_processing_time
        self.one2one_ma_map = one2one_ma_map
        if self.one2one_ma_map:
            assert self.min_ops_per_job == self.max_ops_per_job == self.num_mas

        # determines whether to use a fixed number of total operations or let it vary between instances
        # NOTE: due to the way rl4co builds datasets, we need a fixed size here
        self.n_ops_max = self.max_ops_per_job * self.num_jobs

        # FFSP environment doen't have any other kwargs
        if len(unused_kwargs) > 0:
            log.error(f"Found {len(unused_kwargs)} unused kwargs: {unused_kwargs}")

    def _simulate_processing_times(self, bs, n_ops_max) -> torch.Tensor:
        if self.one2one_ma_map:
            ops_machine_ids = (
                torch.rand((*bs, self.num_jobs, self.num_mas))
                .argsort(dim=-1)
                .flatten(1, 2)
            )
        else:
            ops_machine_ids = torch.randint(
                low=0,
                high=self.num_mas,
                size=(*bs, n_ops_max),
            )
        ops_machine_adj = one_hot(ops_machine_ids, num_classes=self.num_mas)

        # (bs, max_ops, machines)
        proc_times = torch.ones((*bs, n_ops_max, self.num_mas))
        proc_times = torch.randint(
            self.min_processing_time,
            self.max_processing_time + 1,
            size=(*bs, self.num_mas, n_ops_max),
        )

        # remove proc_times for which there is no corresponding ma-ops connection
        proc_times = proc_times * ops_machine_adj.transpose(1, 2)
        # in JSSP there is only one machine capable to process an operation
        assert (proc_times > 0).sum(1).eq(1).all()
        return proc_times.to(torch.float32)

    def _generate(self, batch_size) -> TensorDict:
        # simulate how many operations each job has
        n_ope_per_job = torch.randint(
            self.min_ops_per_job,
            self.max_ops_per_job + 1,
            size=(*batch_size, self.num_jobs),
        )

        # determine the total number of operations per batch instance (which may differ)
        n_ops_batch = n_ope_per_job.sum(1)  # (bs)
        # determine the maximum total number of operations over all batch instances
        n_ops_max = self.n_ops_max or n_ops_batch.max()

        # generate a mask, specifying which operations are padded
        pad_mask = torch.arange(n_ops_max).unsqueeze(0).expand(*batch_size, -1)
        pad_mask = pad_mask.ge(n_ops_batch[:, None].expand_as(pad_mask))

        # determine the id of the end operation for each job
        end_op_per_job = n_ope_per_job.cumsum(1) - 1

        # determine the id of the starting operation for each job
        # (bs, num_jobs)
        start_op_per_job = torch.cat(
            (
                torch.zeros((*batch_size, 1)).to(end_op_per_job),
                end_op_per_job[:, :-1] + 1,
            ),
            dim=1,
        )

        # simulate processing times for machine-operation pairs
        # (bs, num_mas, n_ops_max)
        proc_times = self._simulate_processing_times(batch_size, n_ops_max)

        td = TensorDict(
            {
                "start_op_per_job": start_op_per_job,
                "end_op_per_job": end_op_per_job,
                "proc_times": proc_times,
                "pad_mask": pad_mask,
            },
            batch_size=batch_size,
        )

        return td


class JSSPFileGenerator(Generator):
    """Data generator for the Job-Shop Scheduling Problem (JSSP) using instance files

    Args:
        path: path to files

    Returns:
        A TensorDict with the following key:
            start_op_per_job [batch_size, num_jobs]: first operation of each job
            end_op_per_job [batch_size, num_jobs]: last operation of each job
            proc_times [batch_size, num_machines, total_n_ops]: processing time of ops on machines
            pad_mask [batch_size, total_n_ops]: not all instances have the same number of ops, so padding is used

    """

    def __init__(self, file_path: str, n_ops_max: int = None, **unused_kwargs):
        self.files = (
            [file_path] if os.path.isfile(file_path) else self.list_files(file_path)
        )
        self.num_samples = len(self.files)

        if len(unused_kwargs) > 0:
            log.error(f"Found {len(unused_kwargs)} unused kwargs: {unused_kwargs}")

        if len(self.files) > 1:
            n_ops_max = get_max_ops_from_files(self.files)

        ret = map(partial(read, max_ops=n_ops_max), self.files)

        td_list, num_jobs, num_machines, max_ops_per_job = list(zip(*list(ret)))
        num_jobs, num_machines = map(lambda x: x[0], (num_jobs, num_machines))
        max_ops_per_job = max(max_ops_per_job)

        self.td = torch.cat(td_list, dim=0)
        self.num_mas = num_machines
        self.num_jobs = num_jobs
        self.max_ops_per_job = max_ops_per_job
        self.start_idx = 0

    def _generate(self, batch_size: list[int]) -> TensorDict:
        batch_size = np.prod(batch_size)
        if batch_size > self.num_samples:
            log.warning(
                f"Only found {self.num_samples} instance files, but specified dataset size is {batch_size}"
            )
        end_idx = self.start_idx + batch_size
        td = self.td[self.start_idx : end_idx]
        self.start_idx += batch_size
        if self.start_idx >= self.num_samples:
            self.start_idx = 0
        return td

    @staticmethod
    def list_files(path):
        files = [
            os.path.join(path, f)
            for f in os.listdir(path)
            if os.path.isfile(os.path.join(path, f))
        ]
        assert len(files) > 0, "No files found in the specified path"
        return files



import math

import torch

from tensordict import TensorDict

from rl4co.envs import RL4COEnvBase
from rl4co.utils.decoding import decode_logprobs, process_logits
from rl4co.utils.ops import batchify, unbatchify


def forward_pointer_attn_eas_lay(self, query, key, value, logit_key, mask):
    """Add layer to the forward pass of logit attention, i.e.
    Single-head attention.
    """
    # Compute inner multi-head attention with no projections.
    heads = self._inner_mha(query, key, value, mask)

    # Add residual for EAS layer if is set
    if getattr(self, "eas_layer", None) is not None:
        heads = heads + self.eas_layer(heads)

    glimpse = self.project_out(heads)

    # Batch matrix multiplication to compute logits (batch_size, num_steps, graph_size)
    # bmm is slightly faster than einsum and matmul
    logits = (
        torch.bmm(glimpse, logit_key.squeeze(1).transpose(-2, -1))
        / math.sqrt(glimpse.size(-1))
    ).squeeze(1)

    return logits


def forward_eas(
    self,
    td: TensorDict,
    cached_embeds,
    best_solutions,
    iter_count: int = 0,
    env: str | RL4COEnvBase = None,
    decode_type: str = "multistart_sampling",
    num_starts: int = None,
    mask_logits: bool = True,
    temperature: float = 1.0,
    tanh_clipping: float = 0,
    **decode_kwargs,
):
    """Forward pass of the decoder
    Given the environment state and the pre-computed embeddings, compute the logits and sample actions

    Args:
        td: Input TensorDict containing the environment state
        embeddings: Precomputed embeddings for the nodes. Can be already precomputed cached in form of q, k, v and
        env: Environment to use for decoding. If None, the environment is instantiated from `env_name`. Note that
            it is more efficient to pass an already instantiated environment each time for fine-grained control
        decode_type: Type of decoding to use. Can be one of:
            - "sampling": sample from the logits
            - "greedy": take the argmax of the logits
            - "multistart_sampling": sample as sampling, but with multi-start decoding
            - "multistart_greedy": sample as greedy, but with multi-start decoding
        num_starts: Number of multi-starts to use. If None, will be calculated from the action mask
        calc_reward: Whether to calculate the reward for the decoded sequence
    """
    # TODO: this could be refactored by decoding strategies

    # Collect logprobs
    logprobs = []
    actions = []

    decode_step = 0
    # Multi-start decoding: first action is chosen by ad-hoc node selection
    if num_starts > 1 or "multistart" in decode_type:
        action = env.select_start_nodes(td, num_starts + 1) % num_starts
        # Append incumbent solutions
        if iter_count > 0:
            action = unbatchify(action, num_starts + 1)
            action[:, -1] = best_solutions[:, decode_step]
            action = action.permute(1, 0).reshape(-1)

        # Expand td to batch_size * (num_starts + 1)
        td = batchify(td, num_starts + 1)

        td.set("action", action)
        td = env.step(td)["next"]
        logp = torch.zeros_like(
            td["action_mask"], device=td.device
        )  # first logprobs is 0, so p = logprobs.exp() = 1

        logprobs.append(logp)
        actions.append(action)

    # Main decoding: loop until all sequences are done
    while not td["done"].all():
        decode_step += 1
        logits, mask = self.forward(td, cached_embeds, num_starts + 1)

        logp = process_logits(
            logits,
            mask,
            temperature=self.temperature if self.temperature is not None else temperature,
            tanh_clipping=(
                self.tanh_clipping if self.tanh_clipping is not None else tanh_clipping
            ),
            mask_logits=self.mask_logits if self.mask_logits is not None else mask_logits,
        )

        # Select the indices of the next nodes in the sequences, result (batch_size) long
        action = decode_logprobs(logp, mask, decode_type=decode_type)

        if iter_count > 0:  # append incumbent solutions
            init_shp = action.shape
            action = unbatchify(action, num_starts + 1)
            action[:, -1] = best_solutions[:, decode_step]
            action = action.permute(1, 0).reshape(init_shp)

        td.set("action", action)
        td = env.step(td)["next"]

        # Collect output of step
        logprobs.append(logp)
        actions.append(action)

    logprobs, actions = torch.stack(logprobs, 1), torch.stack(actions, 1)
    rewards = env.get_reward(td, actions)
    return logprobs, actions, td, rewards



import lkh
import numpy as np

from tensordict import TensorDict
from torch import Tensor

from .constants import LKH_SCALING_FACTOR, ROUTEFINDER2LKH
from .utils import scale


def solve(
    instance: TensorDict,
    max_runtime: float,
    problem_type: str,
    num_runs: int,
    solver_loc: str,
) -> tuple[Tensor, Tensor]:
    """
    Solves an AnyVRP instance with OR-Tools.

    Args:
        instance: The AnyVRP instance to solve.
        max_runtime: The maximum runtime for the solver.
        problem_type: The problem type for LKH3.
        num_runs: The number of runs to perform and returns the best result.
        solver_loc: The location of the LKH3 solver executable.

    Returns:
        A tuple containing the action and the cost, respectively.
    """
    problem = instance2problem(instance, problem_type, LKH_SCALING_FACTOR)
    action, cost = _solve(problem, max_runtime, num_runs, solver_loc)
    cost /= -LKH_SCALING_FACTOR

    return action, cost


def _solve(
    problem: lkh.LKHProblem,
    max_runtime: float,
    num_runs: int,
    solver_loc: str,
) -> tuple[Tensor, Tensor]:
    """
    Solves an instance with LKH3.

    Args:
        problem: The LKHProblem instance.
        max_runtime: The maximum runtime for each solver run.
        num_runs: The number of runs to perform and returns the best result.
        solver_loc: The location of the LKH3 solver executable.

    Returns:
        A tuple containing the action and the cost, respectively.
    """
    routes, cost = lkh.solve(
        solver_loc,
        problem=problem,
        time_limit=max_runtime,
        runs=num_runs,
    )

    action = routes2action(routes)
    return action, cost


def instance2problem(
    instance: TensorDict,
    problem_type: str,
    scaling_factor,
) -> lkh.LKHProblem:
    """
    Converts an AnyVRP instance to an LKHProblem instance.

    Args:
        instance: The AnyVRP instance to convert.
        problem_type: The problem type for LKH3.
        scaling_factor: The scaling factor to apply to the instance data.

    Returns:
        The LKHProblem instance.
    """
    num_locations = instance["demand_linehaul"].size()[0]

    # Data specifications
    specs = {}
    specs["DIMENSION"] = num_locations
    specs["CAPACITY"] = scale(instance["vehicle_capacity"], scaling_factor)

    if not np.isinf(distance_limit := instance["distance_limit"]).any():
        specs["DISTANCE"] = scale(distance_limit, scaling_factor)

    specs["EDGE_WEIGHT_TYPE"] = "EXPLICIT"
    specs["EDGE_WEIGHT_FORMAT"] = "FULL_MATRIX"
    specs["NODE_COORD_TYPE"] = "TWOD_COORDS"

    # LKH can only solve VRP variants that are explicitly supported (so no
    # arbitrary combinations between individual supported features). We can
    # support some open variants with some modeling tricks.
    lkh_problem_type = ROUTEFINDER2LKH[problem_type]
    if lkh_problem_type is None:
        raise ValueError(f"Problem type {problem_type} is not supported by LKH.")

    specs["TYPE"] = lkh_problem_type

    # Weird LKH quirk: specifying the number of vehicles lets (D)CVRP hang.
    if lkh_problem_type not in ["CVRP", "DCVRP"]:
        specs["VEHICLES"] = num_locations - 1

    # Data sections
    sections = {}
    sections["NODE_COORD_SECTION"] = scale(instance["locs"], scaling_factor)

    demand_linehaul = scale(instance["demand_linehaul"], scaling_factor)
    demand_backhaul = scale(instance["demand_backhaul"], scaling_factor)
    sections["DEMAND_SECTION"] = demand_linehaul + demand_backhaul

    time_windows = scale(instance["time_windows"], scaling_factor)
    sections["TIME_WINDOW_SECTION"] = time_windows

    service_times = scale(instance["durations"], scaling_factor)
    sections["SERVICE_TIME_SECTION"] = service_times

    distances = instance["cost_matrix"]
    backhaul_class = instance["backhaul_class"]

    if backhaul_class == 1:
        # VRPB has a backhaul section that specifies the backhaul nodes.
        backhaul_idcs = np.flatnonzero(instance["demand_backhaul"]).tolist()
        sections["BACKHAUL_SECTION"] = backhaul_idcs + [-1]

        # linehaul = np.flatnonzero(demand_linehaul > 0)
        # backhaul = np.flatnonzero(demand_backhaul > 0)
        # distances[np.ix_(backhaul, linehaul)] = time_windows.max()

    elif backhaul_class == 2:
        # VRPMPD has a pickup and delivery section that specifies the pickup
        # and delivery quantities for each node, as well as the time windows.
        # The regular time window section is redundant in this case.
        data = [
            [
                0,  # dummy
                time_windows[idx][0],
                time_windows[idx][1],
                service_times[idx],
                demand_backhaul[idx],
                demand_linehaul[idx],
            ]
            for idx in range(num_locations)
        ]
        sections["PICKUP_AND_DELIVERY_SECTION"] = data

    if instance["open_route"]:
        # Arcs to the depot are set to zero as vehicles don’t need to return.
        distances[:, 0] = 0

    sections["EDGE_WEIGHT_SECTION"] = scale(distances, scaling_factor)

    # Convert to VRPLIB-like string.
    problem = "\n".join(f"{k} : {v}" for k, v in specs.items())
    problem += "\n" + "\n".join(_format(name, data) for name, data in sections.items())
    problem += "\n" + "\n".join(["DEPOT_SECTION", "1", "-1", "EOF"])

    return lkh.LKHProblem.parse(problem)


def _is_1D(data) -> bool:
    for elt in data:
        if isinstance(elt, (list, tuple, np.ndarray)):
            return False
    return True


def _format(name: str, data) -> str:
    """
    Formats a data section.

    Args:
        name: The name of the section.
        data: The data to be formatted.

    Returns:
        A VRPLIB-formatted data section.
    """
    section = [name]
    include_idx = name not in ["EDGE_WEIGHT_SECTION", "BACKHAUL_SECTION"]

    if name == "BACKHAUL_SECTION":
        # Treat backhaul section as row vector.
        section.append("\t".join(str(val) for val in data))

    elif _is_1D(data):
        # Treat 1D arrays as column vectors, so each element is a row.
        for idx, elt in enumerate(data, 1):
            prefix = f"{idx}\t" if include_idx else ""
            section.append(prefix + str(elt))
    else:
        for idx, row in enumerate(data, 1):
            prefix = f"{idx}\t" if include_idx else ""
            rest = "\t".join([str(elt) for elt in row])
            section.append(prefix + rest)

    return "\n".join(section)


def routes2action(routes: list[list[int]]) -> list[int]:
    """
    Converts LKH routes to an action.
    """
    # LKH routes are location-indexed, which in turn are 1-indexed. The first
    # location is always the depot, so we subtract 2 to get client indices.
    # LKH routes are 1-indexed, so we subtract 1 to get client indices.
    routes_ = [[client - 1 for client in route] for route in routes]
    return [visit for route in routes_ for visit in route + [0]]



import abc

from typing import Optional, Tuple

import torch
import torch.nn.functional as F

from tensordict.tensordict import TensorDict

from rl4co.envs import RL4COEnvBase
from rl4co.utils.ops import batchify, gather_by_index, unbatchify, unbatchify_and_gather
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def get_decoding_strategy(decoding_strategy, **config):
    strategy_registry = {
        "greedy": Greedy,
        "sampling": Sampling,
        "multistart_greedy": Greedy,
        "multistart_sampling": Sampling,
        "beam_search": BeamSearch,
        "evaluate": Evaluate,
    }

    if decoding_strategy not in strategy_registry:
        log.warning(
            f"Unknown decode type '{decoding_strategy}'. Available decode types: {strategy_registry.keys()}. Defaulting to Sampling."
        )

    if "multistart" in decoding_strategy:
        config["multistart"] = True

    return strategy_registry.get(decoding_strategy, Sampling)(**config)


def get_log_likelihood(logprobs, actions=None, mask=None, return_sum: bool = True):
    """Get log likelihood of selected actions.
    Note that mask is a boolean tensor where True means the value should be kept.

    Args:
        logprobs: Log probabilities of actions from the model (batch_size, seq_len, action_dim).
        actions: Selected actions (batch_size, seq_len).
        mask: Action mask. 1 if feasible, 0 otherwise (so we keep if 1 as done in PyTorch).
        return_sum: Whether to return the sum of log probabilities or not. Defaults to True.
    """
    # Optional: select logp when logp.shape = (bs, dec_steps, N)
    if actions is not None and logprobs.dim() == 3:
        logprobs = logprobs.gather(-1, actions.unsqueeze(-1)).squeeze(-1)

    # Optional: mask out actions irrelevant to objective so they do not get reinforced
    if mask is not None:
        logprobs[~mask] = 0

    assert (
        logprobs > -1000
    ).data.all(), "Logprobs should not be -inf, check sampling procedure!"

    # Calculate log_likelihood
    if return_sum:
        return logprobs.sum(1)  # [batch]
    else:
        return logprobs  # [batch, decode_len]


def decode_logprobs(logprobs, mask, decode_type="sampling"):
    """Decode log probabilities to select actions with mask.
    Note that mask is a boolean tensor where True means the value should be kept.
    """
    if "greedy" in decode_type:
        selected = DecodingStrategy.greedy(logprobs, mask)
    elif "sampling" in decode_type:
        selected = DecodingStrategy.sampling(logprobs, mask)
    else:
        assert False, "Unknown decode type: {}".format(decode_type)
    return selected


def random_policy(td):
    """Helper function to select a random action from available actions"""
    action = torch.multinomial(td["action_mask"].float(), 1).squeeze(-1)
    td.set("action", action)
    return td


def rollout(env, td, policy, max_steps: int = None):
    """Helper function to rollout a policy. Currently, TorchRL does not allow to step
    over envs when done with `env.rollout()`. We need this because for environments that complete at different steps.
    """

    max_steps = float("inf") if max_steps is None else max_steps
    actions = []
    steps = 0

    while not td["done"].all():
        td = policy(td)
        actions.append(td["action"])
        td = env.step(td)["next"]
        steps += 1
        if steps > max_steps:
            log.info("Max steps reached")
            break
    return (
        env.get_reward(td, torch.stack(actions, dim=1)),
        td,
        torch.stack(actions, dim=1),
    )


def modify_logits_for_top_k_filtering(logits, top_k):
    """Set the logits for none top-k values to -inf. Done out-of-place.
    Ref: https://github.com/togethercomputer/stripedhyena/blob/7e13f618027fea9625be1f2d2d94f9a361f6bd02/stripedhyena/sample.py#L6
    """
    indices_to_remove = logits < torch.topk(logits, top_k)[0][..., -1, None]
    return logits.masked_fill(indices_to_remove, float("-inf"))


def modify_logits_for_top_p_filtering(logits, top_p):
    """Set the logits for none top-p values to -inf. Done out-of-place.
    Ref: https://github.com/togethercomputer/stripedhyena/blob/7e13f618027fea9625be1f2d2d94f9a361f6bd02/stripedhyena/sample.py#L14
    """
    if top_p <= 0.0 or top_p >= 1.0:
        return logits

    # First sort and calculate cumulative sum of probabilities.
    sorted_logits, sorted_indices = torch.sort(logits, descending=False)
    cumulative_probs = sorted_logits.softmax(dim=-1).cumsum(dim=-1)

    # Remove tokens with cumulative top_p above the threshold (token with 0 are kept)
    sorted_indices_to_remove = cumulative_probs <= (1 - top_p)

    # Scatter sorted tensors to original indexing
    indices_to_remove = sorted_indices_to_remove.scatter(
        -1, sorted_indices, sorted_indices_to_remove
    )
    return logits.masked_fill(indices_to_remove, float("-inf"))


def process_logits(
    logits: torch.Tensor,
    mask: torch.Tensor = None,
    temperature: float = 1.0,
    top_p: float = 0.0,
    top_k: int = 0,
    tanh_clipping: float = 0,
    mask_logits: bool = True,
):
    """Convert logits to log probabilities with additional features like temperature scaling, top-k and top-p sampling.

    Note:
        We convert to log probabilities instead of probabilities to avoid numerical instability.
        This is because, roughly, softmax = exp(logits) / sum(exp(logits)) and log(softmax) = logits - log(sum(exp(logits))),
        and avoiding the division by the sum of exponentials can help with numerical stability.
        You may check the [official PyTorch documentation](https://pytorch.org/docs/stable/generated/torch.nn.functional.log_softmax.html).

    Args:
        logits: Logits from the model (batch_size, num_actions).
        mask: Action mask. 1 if feasible, 0 otherwise (so we keep if 1 as done in PyTorch).
        temperature: Temperature scaling. Higher values make the distribution more uniform (exploration),
            lower values make it more peaky (exploitation).
        top_p: Top-p sampling, a.k.a. Nucleus Sampling (https://arxiv.org/abs/1904.09751). Remove tokens that have a cumulative probability
            less than the threshold 1 - top_p (lower tail of the distribution). If 0, do not perform.
        top_k: Top-k sampling, i.e. restrict sampling to the top k logits. If 0, do not perform. Note that we only do filtering and
            do not return all the top-k logits here.
        tanh_clipping: Tanh clipping (https://arxiv.org/abs/1611.09940).
        mask_logits: Whether to mask logits of infeasible actions.
    """

    # Tanh clipping from Bello et al. 2016
    if tanh_clipping > 0:
        logits = torch.tanh(logits) * tanh_clipping

    # In RL, we want to mask the logits to prevent the agent from selecting infeasible actions
    if mask_logits:
        assert mask is not None, "mask must be provided if mask_logits is True"
        logits[~mask] = float("-inf")

    logits = logits / temperature  # temperature scaling

    if top_k > 0:
        top_k = min(top_k, logits.size(-1))  # safety check
        logits = modify_logits_for_top_k_filtering(logits, top_k)

    if top_p > 0:
        assert top_p <= 1.0, "top-p should be in (0, 1]."
        logits = modify_logits_for_top_p_filtering(logits, top_p)

    # Compute log probabilities
    return F.log_softmax(logits, dim=-1)


class DecodingStrategy(metaclass=abc.ABCMeta):
    """Base class for decoding strategies. Subclasses should implement the :meth:`_step` method.
    Includes hooks for pre and post main decoding operations.

    Args:
        temperature: Temperature scaling. Higher values make the distribution more uniform (exploration),
            lower values make it more peaky (exploitation). Defaults to 1.0.
        top_p: Top-p sampling, a.k.a. Nucleus Sampling (https://arxiv.org/abs/1904.09751). Defaults to 0.0.
        top_k: Top-k sampling, i.e. restrict sampling to the top k logits. If 0, do not perform. Defaults to 0.
        mask_logits: Whether to mask logits of infeasible actions. Defaults to True.
        tanh_clipping: Tanh clipping (https://arxiv.org/abs/1611.09940). Defaults to 0.
        multisample: Whether to use sampling decoding. Defaults to False.
        num_samples: Number of samples to evaluate during decoding. Defaults to None.
        num_starts: Number of starts for multistart decoding. Defaults to None.
        multistart: Whether to use multistart decoding. Defaults to False.
        select_start_nodes_fn: Function to select start nodes for multistart decoding. Defaults to None.
        improvement_method_mode: Whether to use improvement method mode. Defaults to False.
        select_best: Whether to select the best action or return all. Defaults to False.
        store_all_logp: Whether to store all log probabilities. Defaults to False. If True, logprobs will be stored for all actions.
            Note that this will increase memory usage.
    """

    name = "base"

    def __init__(
        self,
        temperature: float = 1.0,
        top_p: float = 0.0,
        top_k: int = 0,
        mask_logits: bool = True,
        tanh_clipping: float = 0,
        num_samples: Optional[int] = None,
        multisample: bool = False,
        num_starts: Optional[int] = None,
        multistart: bool = False,
        select_start_nodes_fn: Optional[callable] = None,
        improvement_method_mode: bool = False,
        select_best: bool = False,
        store_all_logp: bool = False,
        **kwargs,
    ) -> None:
        self.temperature = temperature
        self.top_p = top_p
        self.top_k = top_k
        self.mask_logits = mask_logits
        self.tanh_clipping = tanh_clipping
        # check if multistart (POMO) and multisample flags
        assert not (
            multistart and multisample
        ), "Using both multistart and multisample is not supported"
        if num_samples and num_starts:
            assert not (
                num_samples > 1 and num_starts > 1
            ), f"num_samples={num_samples} and num_starts={num_starts} are both > 1"
        if num_samples is not None:
            multisample = True if num_samples > 1 else False
        if num_starts is not None:
            multistart = True if num_starts > 1 else False
        self.multistart = multistart
        self.multisample = multisample
        # num_starts is used for both multistart and multisample
        # the function is to use start multiple rollouts for the same instance in parallel
        self.num_starts = num_starts if multistart else num_samples

        self.select_start_nodes_fn = select_start_nodes_fn
        self.improvement_method_mode = improvement_method_mode
        self.select_best = select_best
        self.store_all_logp = store_all_logp
        # initialize buffers
        self.actions = []
        self.logprobs = []

    @abc.abstractmethod
    def _step(
        self,
        logprobs: torch.Tensor,
        mask: torch.Tensor,
        td: Optional[TensorDict] = None,
        action: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> Tuple[torch.Tensor, torch.Tensor, TensorDict]:
        """Main decoding operation. This method should be called in a loop until all sequences are done.

        Args:
            logprobs: Log probabilities processed from logits of the model.
            mask: Action mask. 1 if feasible, 0 otherwise (so we keep if 1 as done in PyTorch).
            td: TensorDict containing the current state of the environment.
            action: Optional action to use, e.g. for evaluating log probabilities.
        """
        raise NotImplementedError("Must be implemented by subclass")

    def pre_decoder_hook(
        self, td: TensorDict, env: RL4COEnvBase, action: Optional[torch.Tensor] = None
    ):
        """Pre decoding hook. This method is called before the main decoding operation."""

        # Multi-start decoding. If num_starts is None, we use the number of actions in the action mask
        if self.multistart or self.multisample:
            if self.num_starts is None:
                self.num_starts = env.get_num_starts(td)
                if self.multisample:
                    log.warning(
                        f"num_starts is not provided for sampling, using num_starts={self.num_starts}"
                    )
        else:
            if self.num_starts is not None:
                if self.num_starts >= 1:
                    log.warning(
                        f"num_starts={self.num_starts} is ignored for decode_type={self.name}"
                    )

            self.num_starts = 0

        # Multi-start decoding: first action is chosen by ad-hoc node selection
        if self.num_starts >= 1:
            if self.multistart:
                if action is None:  # if action is provided, we use it as the first action
                    if self.select_start_nodes_fn is not None:
                        action = self.select_start_nodes_fn(td, env, self.num_starts)
                    else:
                        action = env.select_start_nodes(td, num_starts=self.num_starts)

                # Expand td to batch_size * num_starts
                td = batchify(td, self.num_starts)

                td.set("action", action)
                td = env.step(td)["next"]
                # first logprobs is 0, so p = logprobs.exp() = 1
                if self.store_all_logp:
                    logprobs = torch.zeros_like(td["action_mask"])  # [B, N]
                else:
                    logprobs = torch.zeros_like(action, device=td.device)  # [B]

                self.logprobs.append(logprobs)
                self.actions.append(action)
            else:
                # Expand td to batch_size * num_samplestarts
                td = batchify(td, self.num_starts)

        return td, env, self.num_starts

    def post_decoder_hook(
        self, td: TensorDict, env: RL4COEnvBase
    ) -> Tuple[torch.Tensor, torch.Tensor, TensorDict, RL4COEnvBase]:
        assert (
            len(self.logprobs) > 0
        ), "No logprobs were collected because all environments were done. Check your initial state"
        logprobs = torch.stack(self.logprobs, 1)
        actions = torch.stack(self.actions, 1)
        if self.num_starts > 0 and self.select_best:
            logprobs, actions, td, env = self._select_best(logprobs, actions, td, env)
        return logprobs, actions, td, env

    def step(
        self,
        logits: torch.Tensor,
        mask: torch.Tensor,
        td: Optional[TensorDict] = None,
        action: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> TensorDict:
        """Main decoding operation. This method should be called in a loop until all sequences are done.

        Args:
            logits: Logits from the model.
            mask: Action mask. 1 if feasible, 0 otherwise (so we keep if 1 as done in PyTorch).
            td: TensorDict containing the current state of the environment.
            action: Optional action to use, e.g. for evaluating log probabilities.
        """
        if not self.mask_logits:  # set mask_logit to None if mask_logits is False
            mask = None

        logprobs = process_logits(
            logits,
            mask,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
            tanh_clipping=self.tanh_clipping,
            mask_logits=self.mask_logits,
        )
        logprobs, selected_action, td = self._step(
            logprobs, mask, td, action=action, **kwargs
        )

        # directly return for improvement methods, since the action for improvement methods is finalized in its own policy
        if self.improvement_method_mode:
            return logprobs, selected_action

        # for others
        assert td is not None, "td must be provided"
        if not self.store_all_logp:
            logprobs = gather_by_index(logprobs, selected_action, dim=1)
        td.set("action", selected_action)
        self.actions.append(selected_action)
        self.logprobs.append(logprobs)
        return td

    @staticmethod
    def greedy(logprobs, mask=None):
        """Select the action with the highest probability."""
        # [BS], [BS]
        selected = logprobs.argmax(dim=-1)
        if mask is not None:
            assert (
                not (~mask).gather(1, selected.unsqueeze(-1)).data.any()
            ), "infeasible action selected"

        return selected

    @staticmethod
    def sampling(logprobs, mask=None):
        """Sample an action with a multinomial distribution given by the log probabilities."""
        probs = logprobs.exp()
        selected = torch.multinomial(probs, 1).squeeze(1)

        if mask is not None:
            while (~mask).gather(1, selected.unsqueeze(-1)).data.any():
                log.info("Sampled bad values, resampling!")
                selected = probs.multinomial(1).squeeze(1)
            assert (
                not (~mask).gather(1, selected.unsqueeze(-1)).data.any()
            ), "infeasible action selected"

        return selected

    def _select_best(self, logprobs, actions, td: TensorDict, env: RL4COEnvBase):
        rewards = env.get_reward(td, actions)
        _, max_idxs = unbatchify(rewards, self.num_starts).max(dim=-1)

        actions = unbatchify_and_gather(actions, max_idxs, self.num_starts)
        logprobs = unbatchify_and_gather(logprobs, max_idxs, self.num_starts)
        td = unbatchify_and_gather(td, max_idxs, self.num_starts)

        return logprobs, actions, td, env


class Greedy(DecodingStrategy):
    name = "greedy"

    def _step(
        self, logprobs: torch.Tensor, mask: torch.Tensor, td: TensorDict, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor, TensorDict]:
        """Select the action with the highest log probability"""
        selected = self.greedy(logprobs, mask)
        return logprobs, selected, td


class Sampling(DecodingStrategy):
    name = "sampling"

    def _step(
        self, logprobs: torch.Tensor, mask: torch.Tensor, td: TensorDict, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor, TensorDict]:
        """Sample an action with a multinomial distribution given by the log probabilities."""
        selected = self.sampling(logprobs, mask)
        return logprobs, selected, td


class Evaluate(DecodingStrategy):
    name = "evaluate"

    def _step(
        self,
        logprobs: torch.Tensor,
        mask: torch.Tensor,
        td: TensorDict,
        action: torch.Tensor,
        **kwargs,
    ) -> Tuple[torch.Tensor, torch.Tensor, TensorDict]:
        """The action is provided externally, so we just return the action"""
        selected = action
        return logprobs, selected, td


class BeamSearch(DecodingStrategy):
    name = "beam_search"

    def __init__(self, beam_width=None, select_best=True, **kwargs) -> None:
        # TODO do we really need all logp in beam search?
        kwargs["store_all_logp"] = True
        super().__init__(**kwargs)
        self.beam_width = beam_width
        self.select_best = select_best
        self.parent_beam_logprobs = None
        self.beam_path = []

    def _step(
        self, logprobs: torch.Tensor, mask: torch.Tensor, td: TensorDict, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor, TensorDict]:
        selected, batch_beam_idx = self._make_beam_step(logprobs)
        # select the correct state representation, logprobs and mask according to beam parent
        td = td[batch_beam_idx]
        logprobs = logprobs[batch_beam_idx]
        mask = mask[batch_beam_idx]

        assert (
            not (~mask).gather(1, selected.unsqueeze(-1)).data.any()
        ), "infeasible action selected"

        return logprobs, selected, td

    def pre_decoder_hook(self, td: TensorDict, env: RL4COEnvBase, **kwargs):
        if self.beam_width is None:
            self.beam_width = env.get_num_starts(td)
        assert self.beam_width > 1, "beam width must be larger than 1"

        # select start nodes. TODO: include first step in beam search as well
        if self.select_start_nodes_fn is not None:
            action = self.select_start_nodes_fn(td, env, self.beam_width)
        else:
            action = env.select_start_nodes(td, num_starts=self.beam_width)

        # Expand td to batch_size * beam_width
        td = batchify(td, self.beam_width)

        td.set("action", action)
        td = env.step(td)["next"]

        logprobs = torch.zeros_like(td["action_mask"], device=td.device)
        beam_parent = torch.zeros(logprobs.size(0), device=td.device, dtype=torch.int32)

        self.logprobs.append(logprobs)
        self.actions.append(action)
        self.parent_beam_logprobs = logprobs.gather(1, action[..., None])
        self.beam_path.append(beam_parent)

        return td, env, self.beam_width

    def post_decoder_hook(self, td, env):
        # [BS*BW, seq_len]
        aligned_sequences, aligned_logprobs = self._backtrack()

        if self.select_best:
            return self._select_best_beam(aligned_logprobs, aligned_sequences, td, env)
        else:
            return aligned_logprobs, aligned_sequences, td, env

    def _backtrack(self):
        # [BS*BW, seq_len]
        actions = torch.stack(self.actions, 1)
        # [BS*BW, seq_len]
        logprobs = torch.stack(self.logprobs, 1)
        assert actions.size(1) == len(
            self.beam_path
        ), "action idx shape and beam path shape dont match"

        # [BS*BW]
        cur_parent = self.beam_path[-1]
        # [BS*BW]
        reversed_aligned_sequences = [actions[:, -1]]
        reversed_aligned_logprobs = [logprobs[:, -1]]

        aug_batch_size = actions.size(0)
        batch_size = aug_batch_size // self.beam_width
        batch_beam_sequence = (
            torch.arange(0, batch_size).repeat(self.beam_width).to(actions.device)
        )

        for k in reversed(range(len(self.beam_path) - 1)):
            batch_beam_idx = batch_beam_sequence + cur_parent * batch_size

            reversed_aligned_sequences.append(actions[batch_beam_idx, k])
            reversed_aligned_logprobs.append(logprobs[batch_beam_idx, k])
            cur_parent = self.beam_path[k][batch_beam_idx]

        # [BS*BW, seq_len*num_targets]
        actions = torch.stack(list(reversed(reversed_aligned_sequences)), dim=1)
        logprobs = torch.stack(list(reversed(reversed_aligned_logprobs)), dim=1)

        return actions, logprobs

    def _select_best_beam(self, logprobs, actions, td: TensorDict, env: RL4COEnvBase):
        aug_batch_size = logprobs.size(0)  # num nodes
        batch_size = aug_batch_size // self.beam_width
        rewards = env.get_reward(td, actions)
        _, idx = torch.cat(rewards.unsqueeze(1).split(batch_size), 1).max(1)
        flat_idx = torch.arange(batch_size, device=rewards.device) + idx * batch_size
        return logprobs[flat_idx], actions[flat_idx], td[flat_idx], env

    def _make_beam_step(self, logprobs: torch.Tensor):
        aug_batch_size, num_nodes = logprobs.shape  # num nodes
        batch_size = aug_batch_size // self.beam_width
        batch_beam_sequence = (
            torch.arange(0, batch_size).repeat(self.beam_width).to(logprobs.device)
        )

        # [BS*BW, num_nodes] + [BS*BW, 1] -> [BS*BW, num_nodes]
        log_beam_prob = logprobs + self.parent_beam_logprobs  #

        # [BS, num_nodes * BW]
        log_beam_prob_hstacked = torch.cat(log_beam_prob.split(batch_size), dim=1)
        # [BS, BW]
        topk_logprobs, topk_ind = torch.topk(
            log_beam_prob_hstacked, self.beam_width, dim=1
        )

        # [BS*BW, 1]
        logprobs_selected = torch.hstack(torch.unbind(topk_logprobs, 1)).unsqueeze(1)

        # [BS*BW, 1]
        topk_ind = torch.hstack(torch.unbind(topk_ind, 1))

        # since we stack the logprobs from the distinct branches, the indices in
        # topk dont correspond to node indices directly and need to be translated
        selected = topk_ind % num_nodes  # determine node index

        # calc parent this branch comes from
        beam_parent = (topk_ind // num_nodes).int()

        batch_beam_idx = batch_beam_sequence + beam_parent * batch_size

        self.parent_beam_logprobs = logprobs_selected
        self.beam_path.append(beam_parent)

        return selected, batch_beam_idx



import os

import numpy as np
import numba as nb
from numba import set_num_threads
import torch
from tensordict.tensordict import TensorDict

from rl4co.utils.ops import get_distance_matrix
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def local_search(
    td: TensorDict,
    actions: torch.Tensor,
    max_iterations: int = 1000,
    num_threads: int | None = None,
) -> torch.Tensor:
    """
    Improve the solution using local search, especially 2-opt for TSP.
    Implementation credits to: https://github.com/henry-yeh/DeepACO

    Args:
        td: TensorDict, td from env with shape [batch_size,]
        actions: torch.Tensor, Tour indices with shape [batch_size, num_loc]
        max_iterations: int, maximum number of iterations for 2-opt
    Returns:
        torch.Tensor, Improved tour indices with shape [batch_size, num_loc]
    """
    distances = td.get("distances", None)
    if distances is None:
        distances_np = get_distance_matrix(td["locs"]).numpy()
    else:
        distances_np = distances.detach().cpu().numpy()
    distances_np = distances_np + 1e9 * np.eye(distances_np.shape[1], dtype=np.float32)[None, :, :]

    actions_np = actions.detach().cpu().numpy().astype(np.uint16)
    set_num_threads(num_threads or min(os.cpu_count(), 32))
    numba_results = _two_opt_python(distances_np, actions_np, max_iterations)
    return torch.from_numpy(numba_results.astype(np.int64)).to(actions.device)


@nb.njit(nb.float32(nb.float32[:,:], nb.uint16[:], nb.uint16), nogil=True)
def two_opt_once(distmat, tour, fixed_i = 0):
    '''in-place operation'''
    n = tour.shape[0]
    p = q = 0
    delta = 0
    for i in range(1, n - 1) if fixed_i==0 else range(fixed_i, fixed_i + 1):
        for j in range(i + 1, n):
            node_i, node_j = tour[i], tour[j]
            node_prev, node_next = tour[i - 1], tour[(j + 1) % n]
            if node_prev == node_j or node_next == node_i:
                continue
            change = (
                distmat[node_prev, node_j] + distmat[node_i, node_next]
                - distmat[node_prev, node_i] - distmat[node_j, node_next]
            )
            if change < delta:
                p, q, delta = i, j, change
    if delta < -1e-6:
        tour[p: q + 1] = np.flip(tour[p: q + 1])
        return delta
    else:
        return 0.0

@nb.njit(nb.uint16[:,:](nb.float32[:,:,:], nb.uint16[:,:], nb.int64), nogil=True, parallel=True)
def _two_opt_python(distmat, tour, max_iterations=1000):
    for i in nb.prange(tour.shape[0]):
        iterations = 0
        min_change = -1.0
        while min_change < -1e-6 and iterations < max_iterations:
            min_change = two_opt_once(distmat[i], tour[i], 0)
            iterations += 1
    return tour



import torch
import torch.nn as nn

from tensordict.tensordict import TensorDict

from rl4co.models.nn.ops import PositionalEncoding
from rl4co.utils.ops import batched_scatter_sum, cartesian_to_polar


def env_init_embedding(env_name: str, config: dict) -> nn.Module:
    """Get environment initial embedding. The init embedding is used to initialize the
    general embedding of the problem nodes without any solution information.
    Consists of a linear layer that projects the node features to the embedding space.

    Args:
        env: Environment or its name.
        config: A dictionary of configuration options for the environment.
    """
    embedding_registry = {
        "tsp": TSPInitEmbedding,
        "atsp": TSPInitEmbedding,
        "matnet": MatNetInitEmbedding,
        "cvrp": VRPInitEmbedding,
        "cvrptw": VRPTWInitEmbedding,
        "cvrpmvc": VRPInitEmbedding,
        "svrp": SVRPInitEmbedding,
        "sdvrp": VRPInitEmbedding,
        "pctsp": PCTSPInitEmbedding,
        "spctsp": PCTSPInitEmbedding,
        "op": OPInitEmbedding,
        "dpp": DPPInitEmbedding,
        "mdpp": MDPPInitEmbedding,
        "pdp": PDPInitEmbedding,
        "pdp_ruin_repair": TSPInitEmbedding,
        "tsp_kopt": TSPInitEmbedding,
        "mtsp": MTSPInitEmbedding,
        "smtwtp": SMTWTPInitEmbedding,
        "mdcpdp": MDCPDPInitEmbedding,
        "fjsp": FJSPInitEmbedding,
        "jssp": FJSPInitEmbedding,
        "mtvrp": MTVRPInitEmbedding,
        "shpp": TSPInitEmbedding,
        "flp": FLPInitEmbedding,
        "mcp": MCPInitEmbedding,
    }

    if env_name not in embedding_registry:
        raise ValueError(
            f"Unknown environment name '{env_name}'. Available init embeddings: {embedding_registry.keys()}"
        )

    return embedding_registry[env_name](**config)


class TSPInitEmbedding(nn.Module):
    """Initial embedding for the Traveling Salesman Problems (TSP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the cities
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(TSPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Linear(node_dim, embed_dim, linear_bias)

    def forward(self, td):
        out = self.init_embed(td["locs"])
        return out


class MatNetInitEmbedding(nn.Module):
    """
    Preparing the initial row and column embeddings for MatNet.

    Reference:
    https://github.com/yd-kwon/MatNet/blob/782698b60979effe2e7b61283cca155b7cdb727f/ATSP/ATSP_MatNet/ATSPModel.py#L51


    """

    def __init__(self, embed_dim: int, mode: str = "RandomOneHot") -> None:
        super().__init__()

        self.embed_dim = embed_dim
        assert mode in {
            "RandomOneHot",
            "Random",
        }, "mode must be one of ['RandomOneHot', 'Random']"
        self.mode = mode

    def forward(self, td: TensorDict):
        dmat = td["cost_matrix"]
        b, r, c = dmat.shape

        row_emb = torch.zeros(b, r, self.embed_dim, device=dmat.device)

        if self.mode == "RandomOneHot":
            # MatNet uses one-hot encoding for column embeddings
            # https://github.com/yd-kwon/MatNet/blob/782698b60979effe2e7b61283cca155b7cdb727f/ATSP/ATSP_MatNet/ATSPModel.py#L60
            col_emb = torch.zeros(b, c, self.embed_dim, device=dmat.device)
            rand = torch.rand(b, c)
            rand_idx = rand.argsort(dim=1)
            b_idx = torch.arange(b)[:, None].expand(b, c)
            n_idx = torch.arange(c)[None, :].expand(b, c)
            col_emb[b_idx, n_idx, rand_idx] = 1.0

        elif self.mode == "Random":
            col_emb = torch.rand(b, c, self.embed_dim, device=dmat.device)
        else:
            raise NotImplementedError

        return row_emb, col_emb, dmat


class VRPInitEmbedding(nn.Module):
    """Initial embedding for the Vehicle Routing Problems (VRP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot and customers separately)
        - demand: demand of the customers
    """

    def __init__(self, embed_dim, linear_bias=True, node_dim: int = 3):
        super(VRPInitEmbedding, self).__init__()
        node_dim = node_dim  # 3: x, y, demand
        self.init_embed = nn.Linear(node_dim, embed_dim, linear_bias)
        self.init_embed_depot = nn.Linear(2, embed_dim, linear_bias)  # depot embedding

    def forward(self, td):
        # [batch, 1, 2]-> [batch, 1, embed_dim]
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        depot_embedding = self.init_embed_depot(depot)
        # [batch, n_city, 2, batch, n_city, 1]  -> [batch, n_city, embed_dim]
        node_embeddings = self.init_embed(
            torch.cat((cities, td["demand"][..., None]), -1)
        )
        # [batch, n_city+1, embed_dim]
        out = torch.cat((depot_embedding, node_embeddings), -2)
        return out


class VRPTWInitEmbedding(VRPInitEmbedding):
    def __init__(self, embed_dim, linear_bias=True, node_dim: int = 6):
        # node_dim = 6: x, y, demand, tw start, tw end, service time
        super(VRPTWInitEmbedding, self).__init__(embed_dim, linear_bias, node_dim)

    def forward(self, td):
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        durations = td["durations"][..., 1:]
        time_windows = td["time_windows"][..., 1:, :]
        # embeddings
        depot_embedding = self.init_embed_depot(depot)
        node_embeddings = self.init_embed(
            torch.cat(
                (cities, td["demand"][..., None], time_windows, durations[..., None]), -1
            )
        )
        return torch.cat((depot_embedding, node_embeddings), -2)


class VRPPolarInitEmbedding(nn.Module):
    """Initial embedding for the Vehicle Routing Problems (VRP).
    Embed the following node features to the embedding space, based on polar coordinates:
        - locs: r, theta coordinates of the nodes, with the depot as the origin
        - demand: demand of the customers
    """

    def __init__(
        self,
        embed_dim,
        linear_bias=True,
        node_dim: int = 3,
        attach_cartesian_coords=False,
    ):
        super(VRPPolarInitEmbedding, self).__init__()
        self.node_dim = node_dim + (
            2 if attach_cartesian_coords else 0
        )  # 3: r, theta, demand; 5: r, theta, demand, x, y;
        self.attach_cartesian_coords = attach_cartesian_coords
        self.init_embed = nn.Linear(self.node_dim, embed_dim, linear_bias)
        self.init_embed_depot = nn.Linear(
            self.node_dim, embed_dim, linear_bias
        )  # depot embedding

    def forward(self, td):
        with torch.no_grad():
            locs = td["locs"]
            polar_locs = cartesian_to_polar(locs, locs[..., 0:1, :])
            td["polar_locs"] = polar_locs

            demand = td["demand"]
            demand_with_depot = torch.concat(
                (torch.zeros(demand.shape[0], 1, device=demand.device), demand),
                dim=-1,
            ).unsqueeze(-1)

            if self.attach_cartesian_coords:
                x = torch.concat((polar_locs, demand_with_depot, locs), dim=-1)
            else:
                x = torch.concat((polar_locs, demand_with_depot), dim=-1)

        depot, cities = x[:, :1, :], x[:, 1:, :]
        depot_embedding = self.init_embed_depot(depot)
        node_embeddings = self.init_embed(cities)

        out = torch.cat((depot_embedding, node_embeddings), -2)
        return out


class SVRPInitEmbedding(nn.Module):
    def __init__(self, embed_dim, linear_bias=True, node_dim: int = 3):
        super(SVRPInitEmbedding, self).__init__()
        node_dim = node_dim  # 3: x, y, skill
        self.init_embed = nn.Linear(node_dim, embed_dim, linear_bias)
        self.init_embed_depot = nn.Linear(2, embed_dim, linear_bias)  # depot embedding

    def forward(self, td):
        # [batch, 1, 2]-> [batch, 1, embed_dim]
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        depot_embedding = self.init_embed_depot(depot)
        # [batch, n_city, 2, batch, n_city, 1]  -> [batch, n_city, embed_dim]
        node_embeddings = self.init_embed(torch.cat((cities, td["skills"]), -1))
        # [batch, n_city+1, embed_dim]
        out = torch.cat((depot_embedding, node_embeddings), -2)
        return out


class PCTSPInitEmbedding(nn.Module):
    """Initial embedding for the Prize Collecting Traveling Salesman Problems (PCTSP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot and customers separately)
        - expected_prize: expected prize for visiting the customers.
            In PCTSP, this is the actual prize. In SPCTSP, this is the expected prize.
        - penalty: penalty for not visiting the customers
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(PCTSPInitEmbedding, self).__init__()
        node_dim = 4  # x, y, prize, penalty
        self.init_embed = nn.Linear(node_dim, embed_dim, linear_bias)
        self.init_embed_depot = nn.Linear(2, embed_dim, linear_bias)

    def forward(self, td):
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        depot_embedding = self.init_embed_depot(depot)
        node_embeddings = self.init_embed(
            torch.cat(
                (
                    cities,
                    td["expected_prize"][..., None],
                    td["penalty"][..., 1:, None],
                ),
                -1,
            )
        )
        # batch, n_city+1, embed_dim
        out = torch.cat((depot_embedding, node_embeddings), -2)
        return out


class OPInitEmbedding(nn.Module):
    """Initial embedding for the Orienteering Problems (OP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot and customers separately)
        - prize: prize for visiting the customers
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(OPInitEmbedding, self).__init__()
        node_dim = 3  # x, y, prize
        self.init_embed = nn.Linear(node_dim, embed_dim, linear_bias)
        self.init_embed_depot = nn.Linear(2, embed_dim, linear_bias)  # depot embedding

    def forward(self, td):
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        depot_embedding = self.init_embed_depot(depot)
        node_embeddings = self.init_embed(
            torch.cat(
                (
                    cities,
                    td["prize"][..., 1:, None],  # exclude depot
                ),
                -1,
            )
        )
        out = torch.cat((depot_embedding, node_embeddings), -2)
        return out


class DPPInitEmbedding(nn.Module):
    """Initial embedding for the Decap Placement Problem (DPP), EDA (electronic design automation).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (cells)
        - probe: index of the (single) probe cell. We embed the euclidean distance from the probe to all cells.
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(DPPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Linear(node_dim, embed_dim // 2, linear_bias)  # locs
        self.init_embed_probe = nn.Linear(1, embed_dim // 2, linear_bias)  # probe

    def forward(self, td):
        node_embeddings = self.init_embed(td["locs"])
        probe_embedding = self.init_embed_probe(
            self._distance_probe(td["locs"], td["probe"])
        )
        return torch.cat([node_embeddings, probe_embedding], -1)

    def _distance_probe(self, locs, probe):
        # Euclidean distance from probe to all locations
        probe_loc = torch.gather(locs, 1, probe.unsqueeze(-1).expand(-1, -1, 2))
        return torch.norm(locs - probe_loc, dim=-1).unsqueeze(-1)


class MDPPInitEmbedding(nn.Module):
    """Initial embedding for the Multi-port Placement Problem (MDPP), EDA (electronic design automation).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (cells)
        - probe: indexes of the probe cells (multiple). We embed the euclidean distance of each cell to the closest probe.
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(MDPPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Linear(node_dim, embed_dim, linear_bias)  # locs
        self.init_embed_probe_distance = nn.Linear(
            1, embed_dim, linear_bias
        )  # probe_distance
        self.project_out = nn.Linear(embed_dim * 2, embed_dim, linear_bias)

    def forward(self, td):
        probes = td["probe"]
        locs = td["locs"]
        node_embeddings = self.init_embed(locs)

        # Get the shortest distance from any probe
        dist = torch.cdist(locs, locs, p=2)
        dist[~probes] = float("inf")
        min_dist, _ = torch.min(dist, dim=1)
        min_probe_dist_embedding = self.init_embed_probe_distance(min_dist[..., None])

        return self.project_out(
            torch.cat([node_embeddings, min_probe_dist_embedding], -1)
        )


class PDPInitEmbedding(nn.Module):
    """Initial embedding for the Pickup and Delivery Problem (PDP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot, pickups and deliveries separately)
           Note that pickups and deliveries are interleaved in the input.
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(PDPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed_depot = nn.Linear(2, embed_dim, linear_bias)
        self.init_embed_pick = nn.Linear(node_dim * 2, embed_dim, linear_bias)
        self.init_embed_delivery = nn.Linear(node_dim, embed_dim, linear_bias)

    def forward(self, td):
        depot, locs = td["locs"][..., 0:1, :], td["locs"][..., 1:, :]
        num_locs = locs.size(-2)
        pick_feats = torch.cat(
            [locs[:, : num_locs // 2, :], locs[:, num_locs // 2 :, :]], -1
        )  # [batch_size, graph_size//2, 4]
        delivery_feats = locs[:, num_locs // 2 :, :]  # [batch_size, graph_size//2, 2]
        depot_embeddings = self.init_embed_depot(depot)
        pick_embeddings = self.init_embed_pick(pick_feats)
        delivery_embeddings = self.init_embed_delivery(delivery_feats)
        # concatenate on graph size dimension
        return torch.cat([depot_embeddings, pick_embeddings, delivery_embeddings], -2)


class MTSPInitEmbedding(nn.Module):
    """Initial embedding for the Multiple Traveling Salesman Problem (mTSP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot, cities)
    """

    def __init__(self, embed_dim, linear_bias=True):
        """NOTE: new made by Fede. May need to be checked"""
        super(MTSPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Linear(node_dim, embed_dim, linear_bias)
        self.init_embed_depot = nn.Linear(2, embed_dim, linear_bias)  # depot embedding

    def forward(self, td):
        depot_embedding = self.init_embed_depot(td["locs"][..., 0:1, :])
        node_embedding = self.init_embed(td["locs"][..., 1:, :])
        return torch.cat([depot_embedding, node_embedding], -2)


class SMTWTPInitEmbedding(nn.Module):
    """Initial embedding for the Single Machine Total Weighted Tardiness Problem (SMTWTP).
    Embed the following node features to the embedding space:
        - job_due_time: due time of the jobs
        - job_weight: weights of the jobs
        - job_process_time: the processing time of jobs
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(SMTWTPInitEmbedding, self).__init__()
        node_dim = 3  # job_due_time, job_weight, job_process_time
        self.init_embed = nn.Linear(node_dim, embed_dim, linear_bias)

    def forward(self, td):
        job_due_time = td["job_due_time"]
        job_weight = td["job_weight"]
        job_process_time = td["job_process_time"]
        feat = torch.stack((job_due_time, job_weight, job_process_time), dim=-1)
        out = self.init_embed(feat)
        return out


class MDCPDPInitEmbedding(nn.Module):
    """Initial embedding for the MDCPDP environment
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot, pickups and deliveries separately)
           Note that pickups and deliveries are interleaved in the input.
    """

    def __init__(self, embed_dim, linear_bias=True):
        super(MDCPDPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed_depot = nn.Linear(2, embed_dim, linear_bias)
        self.init_embed_pick = nn.Linear(node_dim * 2, embed_dim, linear_bias)
        self.init_embed_delivery = nn.Linear(node_dim, embed_dim, linear_bias)

    def forward(self, td):
        num_agents = td["capacity"].size(-1)
        depot, locs = td["locs"][..., 0:num_agents, :], td["locs"][..., num_agents:, :]
        num_locs = locs.size(-2)
        pick_feats = torch.cat(
            [locs[:, : num_locs // 2, :], locs[:, num_locs // 2 :, :]], -1
        )  # [batch_size, graph_size//2, 4]
        delivery_feats = locs[:, num_locs // 2 :, :]  # [batch_size, graph_size//2, 2]
        depot_embeddings = self.init_embed_depot(depot)
        pick_embeddings = self.init_embed_pick(pick_feats)
        delivery_embeddings = self.init_embed_delivery(delivery_feats)
        # concatenate on graph size dimension
        return torch.cat([depot_embeddings, pick_embeddings, delivery_embeddings], -2)


class JSSPInitEmbedding(nn.Module):
    def __init__(
        self,
        embed_dim,
        linear_bias: bool = True,
        scaling_factor: int = 1000,
        num_op_feats=5,
    ):
        super(JSSPInitEmbedding, self).__init__()
        self.embed_dim = embed_dim
        self.scaling_factor = scaling_factor
        self.init_ops_embed = nn.Linear(num_op_feats, embed_dim, linear_bias)
        self.pos_encoder = PositionalEncoding(embed_dim, dropout=0.0)

    def _op_features(self, td):
        proc_times = td["proc_times"]
        mean_durations = proc_times.sum(1) / (proc_times.gt(0).sum(1) + 1e-9)
        feats = [
            mean_durations / self.scaling_factor,
            # td["lbs"] / self.scaling_factor,
            td["is_ready"],
            td["num_eligible"],
            td["ops_job_map"],
            td["op_scheduled"],
        ]
        return torch.stack(feats, dim=-1)

    def _init_ops_embed(self, td: TensorDict):
        ops_feat = self._op_features(td)
        ops_emb = self.init_ops_embed(ops_feat)
        ops_emb = self.pos_encoder(ops_emb, td["ops_sequence_order"])

        # zero out padded and finished ops
        mask = td["pad_mask"]  # NOTE dont mask scheduled - leads to instable training
        ops_emb[mask.unsqueeze(-1).expand_as(ops_emb)] = 0
        return ops_emb

    def forward(self, td):
        return self._init_ops_embed(td)


class FJSPInitEmbedding(JSSPInitEmbedding):
    def __init__(self, embed_dim, linear_bias=False, scaling_factor: int = 100):
        super().__init__(embed_dim, linear_bias, scaling_factor)
        self.init_ma_embed = nn.Linear(1, self.embed_dim, bias=linear_bias)
        self.edge_embed = nn.Linear(1, embed_dim, bias=linear_bias)

    def forward(self, td: TensorDict):
        ops_emb = self._init_ops_embed(td)
        ma_emb = self._init_machine_embed(td)
        edge_emb = self._init_edge_embed(td)
        # get edges between operations and machines
        # (bs, ops, ma)
        edges = td["ops_ma_adj"].transpose(1, 2)
        return ops_emb, ma_emb, edge_emb, edges

    def _init_edge_embed(self, td: TensorDict):
        proc_times = td["proc_times"].transpose(1, 2) / self.scaling_factor
        edge_embed = self.edge_embed(proc_times.unsqueeze(-1))
        return edge_embed

    def _init_machine_embed(self, td: TensorDict):
        busy_for = (td["busy_until"] - td["time"].unsqueeze(1)) / self.scaling_factor
        ma_embeddings = self.init_ma_embed(busy_for.unsqueeze(2))
        return ma_embeddings


class FJSPMatNetInitEmbedding(JSSPInitEmbedding):
    def __init__(
        self,
        embed_dim,
        linear_bias: bool = False,
        scaling_factor: int = 1000,
    ):
        super().__init__(embed_dim, linear_bias, scaling_factor)
        self.init_ma_embed = nn.Linear(1, self.embed_dim, bias=linear_bias)

    def _init_machine_embed(self, td: TensorDict):
        busy_for = (td["busy_until"] - td["time"].unsqueeze(1)) / self.scaling_factor
        ma_embeddings = self.init_ma_embed(busy_for.unsqueeze(2))
        return ma_embeddings

    def forward(self, td: TensorDict):
        proc_times = td["proc_times"]
        ops_emb = self._init_ops_embed(td)
        # encoding machines
        ma_emb = self._init_machine_embed(td)
        # edgeweights for matnet
        matnet_edge_weights = proc_times.transpose(1, 2) / self.scaling_factor
        return ops_emb, ma_emb, matnet_edge_weights


class MTVRPInitEmbedding(VRPInitEmbedding):
    def __init__(self, embed_dim, linear_bias=True, node_dim: int = 7):
        # node_dim = 7: x, y, demand_linehaul, demand_backhaul, tw start, tw end, service time
        super(MTVRPInitEmbedding, self).__init__(embed_dim, linear_bias, node_dim)

    def forward(self, td):
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        demand_linehaul, demand_backhaul = (
            td["demand_linehaul"][..., 1:],
            td["demand_backhaul"][..., 1:],
        )
        service_time = td["service_time"][..., 1:]
        time_windows = td["time_windows"][..., 1:, :]
        # [!] convert [0, inf] -> [0, 0] if a problem does not include the time window constraint, do not modify in-place
        time_windows = torch.nan_to_num(time_windows, posinf=0.0)
        # embeddings
        depot_embedding = self.init_embed_depot(depot)
        node_embeddings = self.init_embed(
            torch.cat(
                (
                    cities,
                    demand_linehaul[..., None],
                    demand_backhaul[..., None],
                    time_windows,
                    service_time[..., None],
                ),
                -1,
            )
        )
        return torch.cat((depot_embedding, node_embeddings), -2)


class FLPInitEmbedding(nn.Module):
    def __init__(self, embed_dim: int):
        super().__init__()
        self.projection = nn.Linear(2, embed_dim, bias=True)

    def forward(self, td: TensorDict):
        hdim = self.projection(td["locs"])
        return hdim


class MCPInitEmbedding(nn.Module):
    def __init__(self, embed_dim: int):
        super().__init__()
        self.projection_items = nn.Linear(1, embed_dim, bias=True)

    def forward(self, td: TensorDict):
        items_embed = self.projection_items(td["weights"].unsqueeze(-1))
        # sum pooling
        membership_emb = batched_scatter_sum(items_embed, td["membership"].long())
        return membership_emb



import torch


def render(td, actions=None, ax=None):
    import matplotlib.pyplot as plt
    import numpy as np

    from matplotlib import colormaps

    # Create a plot of the nodes
    if ax is None:
        _, ax = plt.subplots()

    td = td.detach().cpu()

    # Actions
    if actions is None:
        actions = td.get("action", None)
    actions = actions.detach().cpu() if actions is not None else None

    # if batch_size greater than 0 , we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]
        actions = actions[0] if actions is not None else None

    # Variables
    depot = td["locs"][0, :]
    customers = td["locs"][1:, :]
    prizes = td["real_prize"][1:]
    penalties = td["penalty"][1:]
    normalized_prizes = (
        200 * (prizes - torch.min(prizes)) / (torch.max(prizes) - torch.min(prizes)) + 10
    )
    normalized_penalties = (
        3
        * (penalties - torch.min(penalties))
        / (torch.max(penalties) - torch.min(penalties))
    )

    # Represent penalty with colormap and size of edges
    penalty_cmap = colormaps.get_cmap("BuPu")
    penalty_colors = penalty_cmap(normalized_penalties)

    # Plot depot and customers with prize (size of nodes) and penalties (size of borders)
    ax.scatter(
        depot[0],
        depot[1],
        marker="s",
        c="tab:green",
        edgecolors="black",
        zorder=1,
        s=100,
    )  # Plot depot as square
    ax.scatter(
        customers[:, 0],
        customers[:, 1],
        s=normalized_prizes,
        c=normalized_prizes,
        cmap="autumn_r",
        alpha=1,
        edgecolors=penalty_colors,
        linewidths=normalized_penalties,
    )  # Plot all customers with size and color indicating the prize

    # Gather locs in order of action if available
    if actions is None:
        print("No action in TensorDict, rendering unsorted locs")
    else:
        # Reorder the customers and their corresponding prizes based on actions
        tour = customers[actions - 1]  # subtract 1 to match Python's 0-indexing

        # Append the depot at the beginning and the end of the tour
        tour = np.vstack((depot, tour, depot))

        # Use quiver to plot the tour
        dx, dy = np.diff(tour[:, 0]), np.diff(tour[:, 1])
        ax.quiver(
            tour[:-1, 0],
            tour[:-1, 1],
            dx,
            dy,
            scale_units="xy",
            angles="xy",
            scale=1,
            zorder=2,
            color="black",
            width=0.0035,
        )

    return ax



import time

from functools import partial
from typing import Any

import torch

from lightning.pytorch.utilities.types import STEP_OUTPUT
from torch.utils.data import Dataset

from rl4co.data.transforms import StateAugmentation
from rl4co.models.common.transductive import TransductiveModel
from rl4co.utils.ops import batchify, unbatchify
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class ActiveSearch(TransductiveModel):
    """Active Search for Neural Combination Optimization from Bello et al. (2016).
    Fine-tunes the whole policy network (encoder + decoder) on a batch of instances.
    Reference: https://arxiv.org/abs/1611.09940

    Args:
        env: RL4CO environment to be solved
        policy: policy network
        dataset: dataset to be used for training
        batch_size: batch size for training
        max_iters: maximum number of iterations
        augment_size: number of augmentations per state
        augment_dihedral: whether to augment with dihedral rotations
        parallel_runs: number of parallel runs
        max_runtime: maximum runtime in seconds
        save_path: path to save solution checkpoints
        optimizer: optimizer to use for training
        optimizer_kwargs: keyword arguments for optimizer
        **kwargs: additional keyword arguments
    """

    def __init__(
        self,
        env,
        policy,
        dataset: Dataset | str,
        batch_size: int = 1,
        max_iters: int = 200,
        augment_size: int = 8,
        augment_dihedral: bool = True,
        num_parallel_runs: int = 1,
        max_runtime: int = 86_400,
        save_path: str = None,
        optimizer: str | torch.optim.Optimizer | partial = "Adam",
        optimizer_kwargs: dict = {"lr": 2.6e-4, "weight_decay": 1e-6},
        **kwargs,
    ):
        self.save_hyperparameters(logger=False)

        assert batch_size == 1, "Batch size must be 1 for active search"

        super(ActiveSearch, self).__init__(
            env,
            policy=policy,
            dataset=dataset,
            batch_size=batch_size,
            max_iters=max_iters,
            max_runtime=max_runtime,
            save_path=save_path,
            optimizer=optimizer,
            optimizer_kwargs=optimizer_kwargs,
            **kwargs,
        )

    def setup(self, stage="fit"):
        """Setup base class and instantiate:
        - augmentation
        - instance solutions and rewards
        - original policy state dict
        """
        log.info("Setting up active search...")
        super(ActiveSearch, self).setup(stage)

        # Instantiate augmentation
        self.augmentation = StateAugmentation(
            num_augment=self.hparams.augment_size,
            augment_fn="dihedral8" if self.hparams.augment_dihedral else "symmetric",
        )

        # Store original policy state dict
        self.original_policy_state = self.policy.state_dict()

        # Get dataset size and problem size
        dataset_size = len(self.dataset)
        _batch = next(iter(self.train_dataloader()))
        self.problem_size = self.env.reset(_batch)["action_mask"].shape[-1]
        self.instance_solutions = torch.zeros(
            dataset_size, self.problem_size * 2, dtype=int
        )
        self.instance_rewards = torch.zeros(dataset_size)

    def on_train_batch_start(self, batch: Any, batch_idx: int):
        """Called before training (i.e. search) for a new batch begins.
        We re-load the original policy state dict and configure the optimizer.
        """
        self.policy.load_state_dict(self.original_policy_state)
        self.configure_optimizers(self.policy.parameters())

    def training_step(self, batch, batch_idx):
        """Main search loop. We use the training step to effectively adapt to a `batch` of instances."""
        # Augment state
        batch_size = batch.shape[0]
        td_init = self.env.reset(batch)
        n_aug, n_start, n_runs = (
            self.augmentation.num_augment,
            self.env.get_num_starts(td_init),
            self.hparams.num_parallel_runs,
        )
        td_init = self.augmentation(td_init)
        td_init = batchify(td_init, n_runs)

        # Solution and reward buffer
        max_reward = torch.full((batch_size,), -float("inf"), device=batch.device)
        best_solutions = torch.zeros(
            batch_size, self.problem_size * 2, device=batch.device, dtype=int
        )

        # Init search
        t_start = time.time()
        for i in range(self.hparams.max_iters):
            # Evaluate policy with sampling multistarts (as in POMO)
            out = self.policy(
                td_init.clone(),
                env=self.env,
                decode_type="multistart_sampling",
                num_starts=n_start,
            )

            if i == 0:
                log.info(f"Initial reward: {out['reward'].max():.2f}")

            # Update best solution and reward found
            max_reward_iter = out["reward"].max()
            if max_reward_iter > max_reward:
                max_reward_idx = out["reward"].argmax()
                best_solution_iter = out["actions"][max_reward_idx]
                max_reward = max_reward_iter
                best_solutions[0, : best_solution_iter.shape[0]] = best_solution_iter

            # Compute REINFORCE loss with shared baseline
            reward = unbatchify(out["reward"], (n_runs, n_aug, n_start))
            ll = unbatchify(out["log_likelihood"], (n_runs, n_aug, n_start))
            advantage = reward - reward.mean(dim=-1, keepdim=True)
            loss = -(advantage * ll).mean()

            # Backpropagate loss
            # perform manual optimization following the Lightning routine
            # https://lightning.ai/docs/pytorch/stable/common/optimization.html
            opt = self.optimizers()
            opt.zero_grad()
            self.manual_backward(loss)

            self.log_dict(
                {
                    "loss": loss,
                    "max_reward": max_reward,
                    "step": i,
                    "time": time.time() - t_start,
                },
                on_step=self.log_on_step,
            )

            # Stop if max runtime is exceeded
            if time.time() - t_start > self.hparams.max_runtime:
                break

        return {"max_reward": max_reward, "best_solutions": best_solutions}

    def on_train_batch_end(
        self, outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        """We store the best solution and reward found."""
        max_rewards, best_solutions = outputs["max_reward"], outputs["best_solutions"]
        self.instance_rewards[batch_idx] = max_rewards
        self.instance_solutions[batch_idx, :] = best_solutions.squeeze(
            0
        )  # only one instance
        log.info(f"Best reward: {max_rewards.mean():.2f}")

    def on_train_epoch_end(self) -> None:
        """Called when the training ends.
        If the epoch ends, it means we have finished searching over the
        instances, thus the trainer should stop.
        """
        save_path = self.hparams.save_path
        if save_path is not None:
            log.info(f"Saving solutions and rewards to {save_path}...")
            torch.save(
                {"solutions": self.instance_solutions, "rewards": self.instance_rewards},
                save_path,
            )

        # https://github.com/Lightning-AI/lightning/issues/1406
        self.trainer.should_stop = True



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

from .generator import MPDPGenerator
from .render import render

log = get_pylogger(__name__)


class MPDPEnv(RL4COEnvBase):
    """Multi-agent Pickup and Delivery Problem (mPDP) environment.
    The goal is to pick up and deliver all the packages while satisfying the precedence constraints.
    When an agent goes back to the depot, a new agent is spawned. In the min-max version, the goal is to minimize the
    maximum tour length among all agents. The reward is 0 unless the agent visits all the customers.
    In that case, the reward is (-)length of the path: maximizing the reward is equivalent to minimizing the path length.

    Observations:
        - locations of the depot, pickup, and delivery locations
        - current location of the vehicle
        - the remaining locations to deliver
        - the visited locations
        - the current step

    Constraints:
        - the tour starts and ends at the depot
        - each pickup location must be visited before its corresponding delivery location
        - the vehicle cannot visit the same location twice

    Finish Condition:
        - the vehicle has visited all locations

    Reward:
        - (minus) the negative length of the path

    Args:
        generator: MPDPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "mpdp"

    def __init__(
        self,
        generator: MPDPGenerator = None,
        generator_params: dict = {},
        objective: str = "minmax",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = MPDPGenerator(**generator_params)
        self.generator = generator
        self.objective = objective
        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        selected = td["action"][:, None]  # Add dimension for step

        agent_num = td["lengths"].size(1)
        n_loc = td["to_delivery"].size(-1) - agent_num - 1

        new_to_delivery = (selected + n_loc // 2) % (
            n_loc + agent_num + 1
        )  # the pair node of selected node

        is_request = (selected > agent_num) & (selected <= agent_num + n_loc // 2)
        td["left_request"][is_request] -= 1
        depot_distance = td["depot_distance"].scatter(-1, selected, 0)

        add_pd = td["add_pd_distance"][is_request.squeeze(-1), :].gather(
            -1, selected[is_request.squeeze(-1), :] - agent_num - 1
        )
        td["longest_lengths"][is_request.squeeze(-1), :].scatter_add_(
            -1, td["count_depot"][is_request.squeeze(-1), :], add_pd
        )
        td["add_pd_distance"][is_request.squeeze(-1), :].scatter_(
            -1, selected[is_request.squeeze(-1), :] - agent_num - 1, 0
        )
        remain_sum_paired_distance = td["add_pd_distance"].sum(-1, keepdim=True)
        remain_pickup_max_distance = depot_distance[:, : agent_num + 1 + n_loc // 2].max(
            dim=-1, keepdim=True
        )[0]
        remain_delivery_max_distance = depot_distance[
            :, agent_num + 1 + n_loc // 2 :
        ].max(dim=-1, keepdim=True)[0]

        # Calculate makespan
        cur_coord = gather_by_index(td["locs"], selected)
        path_lengths = (cur_coord - td["cur_coord"]).norm(p=2, dim=-1)

        td["lengths"].scatter_add_(-1, td["count_depot"], path_lengths.unsqueeze(-1))

        # If visit depot then plus one to count_depot\
        td["count_depot"][
            (selected == td["agent_idx"]) & (td["agent_idx"] < agent_num)
        ] += 1  # torch.ones(td["count_depot"][(selected == 0) & (td["agent_idx"] < agent_num)].shape, dtype=torch.int64, device=td["count_depot"].device)

        # `agent_idx` is added by 1 if the current agent comes back to depot
        agent_idx = (td["count_depot"] + 1) * torch.ones(
            selected.size(0), 1, dtype=torch.long, device=td["count_depot"].device
        )
        visited = td["visited"].scatter(-1, selected.unsqueeze(-1), 1)
        to_delivery = td["to_delivery"].scatter(-1, new_to_delivery[:, :, None], 1)

        # Get done and reward
        done = visited.all(dim=-1, keepdim=True).squeeze(-1)
        reward = torch.zeros_like(done)

        td.update(
            {
                "visited": visited,
                "agent_idx": agent_idx,
                "cur_coord": cur_coord,
                "to_delivery": to_delivery,
                "depot_distance": depot_distance,
                "remain_sum_paired_distance": remain_sum_paired_distance,
                "remain_pickup_max_distance": remain_pickup_max_distance,
                "remain_delivery_max_distance": remain_delivery_max_distance,
                "i": td["i"] + 1,
                "done": done,
                "reward": reward,
            }
        )
        td.set("action_mask", self.get_action_mask(td))
        return td

    def _reset(
        self,
        td: Optional[TensorDict] = None,
        batch_size: Optional[list] = None,
        agent_num: Optional[int] = None,  # NOTE hardcoded from ET
    ) -> TensorDict:
        device = td.device

        # NOTE: this is a hack to get the agent_num
        # agent_num = td["agent_num"][0].item() if agent_num is None else agent_num
        # agent_num = agent_num if agent_num is not None else td["agent_num"][0].item()

        depot = td["depot"]
        depot = depot.repeat(1, agent_num + 1, 1)
        loc = td["locs"]
        left_request = loc.size(1) // 2
        whole_instance = torch.cat((depot, loc), dim=1)

        # Distance from all nodes between each other
        distance = torch.cdist(whole_instance, whole_instance, p=2)
        index = torch.arange(left_request, 2 * left_request, device=device)[None, :, None]
        index = index.repeat(distance.shape[0], 1, 1)
        add_pd_distance = distance[
            :, agent_num + 1 : agent_num + 1 + left_request, agent_num + 1 :
        ].gather(-1, index)
        add_pd_distance = add_pd_distance.squeeze(-1)

        remain_pickup_max_distance = distance[:, 0, : agent_num + 1 + left_request].max(
            dim=-1, keepdim=True
        )[0]
        remain_delivery_max_distance = distance[:, 0, agent_num + 1 + left_request :].max(
            dim=-1, keepdim=True
        )[0]
        remain_sum_paired_distance = add_pd_distance.sum(dim=-1, keepdim=True)

        # Distance from depot to all nodes
        # Delivery nodes should consider the sum of distance from depot to paired pickup nodes and pickup nodes to delivery nodes
        distance[:, 0, agent_num + 1 : agent_num + 1 + left_request] = (
            distance[:, 0, agent_num + 1 : agent_num + 1 + left_request]
            + distance[:, 0, agent_num + 1 + left_request :]
        )

        # Distance from depot to all nodes
        depot_distance = distance[:, 0, :]
        depot_distance[:, agent_num + 1 : agent_num + 1 + left_request] = depot_distance[
            :, agent_num + 1 : agent_num + 1 + left_request
        ]  # + add_pd_distance

        batch_size, n_loc, _ = loc.size()
        to_delivery = torch.cat(
            [
                torch.ones(
                    batch_size,
                    1,
                    n_loc // 2 + agent_num + 1,
                    dtype=torch.uint8,
                    device=device,
                ),
                torch.zeros(batch_size, 1, n_loc // 2, dtype=torch.uint8, device=device),
            ],
            dim=-1,
        )

        # Create reset TensorDict
        td_reset = TensorDict(
            {
                "locs": torch.cat((depot, loc), -2),
                "visited": torch.zeros(
                    batch_size,
                    1,
                    n_loc + agent_num + 1,
                    dtype=torch.uint8,
                    device=device,
                ),
                "lengths": torch.zeros(batch_size, agent_num, device=device),
                "longest_lengths": torch.zeros(batch_size, agent_num, device=device),
                "cur_coord": (
                    td["depot"] if len(td["depot"].shape) == 2 else td["depot"].squeeze(1)
                ),
                "i": torch.zeros(
                    batch_size, dtype=torch.int64, device=device
                ),  # Vector with length num_steps
                "to_delivery": to_delivery,
                "count_depot": torch.zeros(
                    batch_size, 1, dtype=torch.int64, device=device
                ),
                "agent_idx": torch.ones(batch_size, 1, dtype=torch.long, device=device),
                "left_request": left_request
                * torch.ones(batch_size, 1, dtype=torch.long, device=device),
                "remain_pickup_max_distance": remain_pickup_max_distance,
                "remain_delivery_max_distance": remain_delivery_max_distance,
                "depot_distance": depot_distance,
                "remain_sum_paired_distance": remain_sum_paired_distance,
                "add_pd_distance": add_pd_distance,
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        """Get the action mask for the current state."""

        visited_loc = td["visited"].clone()

        agent_num = td["lengths"].size(1)
        n_loc = visited_loc.size(-1) - agent_num - 1  # num of customers
        batch_size = visited_loc.size(0)
        agent_idx = td["agent_idx"][:, None, :]
        mask_loc = visited_loc.to(td["to_delivery"].device) | (1 - td["to_delivery"])

        # depot
        if td["i"][0].item() != 0:
            mask_loc[:, :, : agent_num + 1] = 1

            # if deliver nodes which is assigned agent is complete, then agent can go to depot
            no_item_to_delivery = (
                visited_loc[:, :, n_loc // 2 + agent_num + 1 :]
                == td["to_delivery"][:, :, n_loc // 2 + agent_num + 1 :]
            ).all(dim=-1)
            mask_loc[no_item_to_delivery.squeeze(-1), :, :] = mask_loc[
                no_item_to_delivery.squeeze(-1), :, :
            ].scatter_(-1, agent_idx[no_item_to_delivery.squeeze(-1), :, :], 0)

            condition = (td["count_depot"] == agent_num - 1) & (
                (visited_loc[:, :, agent_num + 1 :] == 0).sum(dim=-1) != 0
            )

            mask_loc[..., agent_num][condition] = 1

        else:
            return (
                torch.cat(
                    [
                        torch.zeros(
                            batch_size, 1, 1, dtype=torch.uint8, device=mask_loc.device
                        ),
                        torch.ones(
                            batch_size,
                            1,
                            n_loc + agent_num,
                            dtype=torch.uint8,
                            device=mask_loc.device,
                        ),
                    ],
                    dim=-1,
                )
                > 0
            )
        action_mask = mask_loc == 0  # action_mask gets feasible actions
        return action_mask

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        # Calculate the reward (negative tour length)
        if self.objective == "minmax":
            return -td["lengths"].max(dim=-1, keepdim=True)[0].squeeze(-1)
        elif self.objective == "minsum":
            return -td["lengths"].sum(dim=-1, keepdim=True).squeeze(-1)
        else:
            raise ValueError(f"Unknown objective {self.objective}")

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor) -> None:
        assert True, "Not implemented"

    def _make_spec(self, generator: MPDPGenerator):
        """Make the observation and action specs from the parameters."""
        max_nodes = self.num_loc + self.max_num_agents + 1
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(max_nodes, 2),
                dtype=torch.float32,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(max_nodes, 1),
                dtype=torch.bool,
            ),
            visited=Unbounded(
                shape=(1, max_nodes),
                dtype=torch.bool,
            ),
            lengths=Unbounded(
                shape=(generator.max_num_agents,),
                dtype=torch.float32,
            ),
            longest_lengths=Unbounded(
                shape=(generator.max_num_agents,),
                dtype=torch.float32,
            ),
            cur_coord=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(2,),
                dtype=torch.float32,
            ),
            to_delivery=Unbounded(
                shape=(max_nodes, 1),
                dtype=torch.bool,
            ),
            count_depot=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            agent_idx=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            left_request=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
            remain_pickup_max_distance=Unbounded(
                shape=(1,),
                dtype=torch.float32,
            ),
            remain_delivery_max_distance=Unbounded(
                shape=(1,),
                dtype=torch.float32,
            ),
            depot_distance=Unbounded(
                shape=(max_nodes,),
                dtype=torch.float32,
            ),
            remain_sum_paired_distance=Unbounded(
                shape=(1,),
                dtype=torch.float32,
            ),
            add_pd_distance=Unbounded(
                shape=(max_nodes,),
                dtype=torch.float32,
            ),
            ## NOTE: we should have a vectorized implementation for agent_num
            # agent_num=Unbounded(
            #     shape=(1,),
            #     dtype=torch.int64,
            # ),
            i=Unbounded(
                shape=(1,),
                dtype=torch.int64,
            ),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=max_nodes,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)



from typing import Optional

import torch

from tensordict.tensordict import TensorDict

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

from .generator import FLPGenerator

log = get_pylogger(__name__)


class FLPEnv(RL4COEnvBase):
    """Facility Location Problem (FLP) environment
    At each step, the agent chooses a location. The reward is 0 unless enough number of locations are chosen.
    The reward is (-) the total distance of each location to its closest chosen location.

    Observations:
        - the locations
        - the number of locations to choose

    Constraints:
        - the given number of locations must be chosen

    Finish condition:
        - the given number of locations are chosen

    Reward:
        - (minus) the total distance of each location to its closest chosen location

    Args:
        generator: FLPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "flp"

    def __init__(
        self,
        generator: FLPGenerator = None,
        generator_params: dict = {},
        check_solution=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = FLPGenerator(**generator_params)
        self.generator = generator
        self.check_solution = check_solution
        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        # action: [batch_size, 1]; the location to be chosen in each instance
        selected = td["action"]
        batch_size = selected.shape[0]

        # Update location selection status
        chosen = td["chosen"].clone()  # (batch_size, n_locations)
        n_points_ = chosen.shape[-1]

        chosen[torch.arange(batch_size).to(td.device), selected] = True

        # We are done if we choose enough locations
        done = td["i"] >= (td["to_choose"] - 1)

        # The reward is calculated outside via get_reward for efficiency, so we set it to zero here
        reward = torch.zeros_like(done)

        # Update distances
        orig_distances = td["orig_distances"]  # (batch_size, n_points, n_points)

        cur_min_dist = (
            gather_by_index(
                orig_distances, chosen.nonzero(as_tuple=True)[1].view(batch_size, -1)
            )
            .view(batch_size, -1, n_points_)
            .min(dim=1)
            .values
        )

        # We cannot choose the already-chosen locations
        action_mask = ~chosen

        td.update(
            {
                "distances": cur_min_dist,  # (batch_size, n_points)
                # states changed by actions
                "chosen": chosen,  # each entry is binary; 1 iff the corresponding facility is chosen
                "i": td["i"] + 1,  # the number of sets we have chosen
                "action_mask": action_mask,
                "reward": reward,
                "done": done,
            }
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        self.to(td.device)

        return TensorDict(
            {
                # given information
                "locs": td["locs"],  # (batch_size, n_points, dim_loc)
                "orig_distances": td[
                    "orig_distances"
                ],  # (batch_size, n_points, n_points)
                "distances": td["distances"],  # (batch_size, n_points, n_points)
                # states changed by actions
                "chosen": torch.zeros(
                    *td["locs"].shape[:-1], dtype=torch.bool, device=td.device
                ),  # each entry is binary; 1 iff the corresponding facility is chosen
                "to_choose": td["to_choose"],  # the number of sets to choose
                "i": torch.zeros(
                    *batch_size, dtype=torch.int64, device=td.device
                ),  # the number of sets we have chosen
                "action_mask": torch.ones(
                    *td["locs"].shape[:-1], dtype=torch.bool, device=td.device
                ),
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: FLPGenerator):
        # TODO: make spec
        pass

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        if self.check_solution:
            self.check_solution_validity(td, actions)

        # The reward is (minus) the total distance from each location to the closest chosen location
        chosen = td["chosen"]  # (batch_size, n_points)
        batch_size_ = td["chosen"].shape[0]
        n_points_ = td["chosen"].shape[-1]
        orig_distances = td["orig_distances"]
        cur_min_dist = (
            gather_by_index(
                orig_distances, chosen.nonzero(as_tuple=True)[1].view(batch_size_, -1)
            )
            .view(batch_size_, -1, n_points_)
            .min(1)
            .values.sum(-1)
        )
        return -cur_min_dist

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor) -> None:
        # TODO: check solution validity
        pass

    @staticmethod
    def local_search(td: TensorDict, actions: torch.Tensor, **kwargs) -> torch.Tensor:
        # TODO: local search
        pass

    @staticmethod
    def get_num_starts(td):
        return td["action_mask"].shape[-1]

    @staticmethod
    def select_start_nodes(td, num_starts):
        num_loc = td["action_mask"].shape[-1]
        return (
            torch.arange(num_starts, device=td.device).repeat_interleave(td.shape[0])
            % num_loc
        )



from functools import lru_cache

import torch

from tensordict import TensorDict

from rl4co.models.common.constructive.base import ConstructiveDecoder
from rl4co.utils.ops import batchify


@lru_cache(10)
def _multistart_batched_index(batch_size: int, num_starts: int):
    """Create a batched index for multistart decoding"""
    arr = torch.arange(batch_size)
    if num_starts <= 1:
        return arr
    else:
        return batchify(arr, num_starts)


class NonAutoregressiveDecoder(ConstructiveDecoder):
    """The nonautoregressive decoder is a simple callable class that
    takes the tensor dictionary and the heatmaps logits and returns the logits for the current
    action logits and the action mask.
    """

    def forward(self, td: TensorDict, heatmaps_logits: torch.Tensor, num_starts: int):
        return self.heatmap_to_logits(td, heatmaps_logits, num_starts)

    @staticmethod
    def heatmap_to_logits(td: TensorDict, heatmaps_logits: torch.Tensor, num_starts: int):
        """Obtain heatmap logits for current action to the next ones"""
        batch_size = heatmaps_logits.shape[0]
        _indexer = _multistart_batched_index(batch_size, num_starts)
        assert _indexer.shape[0] == td.shape[0]

        current_node = td.get("current_node", None).squeeze(-1)
        if current_node is None:
            logits = heatmaps_logits.mean(-1)
            logits = logits[_indexer, :]
        else:
            logits = heatmaps_logits[_indexer, current_node, :]
        return logits, td["action_mask"]



from typing import Callable, Tuple

import torch.nn as nn
import torch.nn.functional as F

from tensordict import TensorDict
from torch import Tensor

try:
    from torch_geometric.nn import GCNConv
except ImportError:
    GCNConv = None
from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.utils.ops import get_full_graph_edge_index
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


EdgeIndexFnSignature = Callable[[TensorDict, int], Tensor]


def edge_idx_fn_wrapper(td: TensorDict, num_nodes: int):
    # self-loop is added by GCNConv layer
    return get_full_graph_edge_index(num_nodes, self_loop=False).to(td.device)


class GCNEncoder(nn.Module):
    """Graph Convolutional Network to encode embeddings with a series of GCN
    layers from the pytorch geometric package

    Args:
        embed_dim: dimension of the embeddings
        num_nodes: number of nodes in the graph
        num_gcn_layer: number of GCN layers
        self_loop: whether to add self loop in the graph
        residual: whether to use residual connection
    """

    def __init__(
        self,
        env_name: str,
        embed_dim: int,
        num_layers: int,
        init_embedding: nn.Module = None,
        residual: bool = True,
        edge_idx_fn: EdgeIndexFnSignature = None,
        dropout: float = 0.5,
        bias: bool = True,
    ):
        super().__init__()

        self.env_name = env_name
        self.embed_dim = embed_dim
        self.residual = residual
        self.dropout = dropout

        self.init_embedding = (
            env_init_embedding(self.env_name, {"embed_dim": embed_dim})
            if init_embedding is None
            else init_embedding
        )

        if edge_idx_fn is None:
            log.warning("No edge indices passed. Assume a fully connected graph")
            edge_idx_fn = edge_idx_fn_wrapper

        self.edge_idx_fn = edge_idx_fn

        # Define the GCN layers
        self.gcn_layers = nn.ModuleList(
            [GCNConv(embed_dim, embed_dim, bias=bias) for _ in range(num_layers)]
        )

    def forward(
        self, td: TensorDict, mask: Tensor | None = None
    ) -> Tuple[Tensor, Tensor]:
        """Forward pass of the encoder.
        Transform the input TensorDict into a latent representation.

        Args:
            td: Input TensorDict containing the environment state
            mask: Mask to apply to the attention

        Returns:
            h: Latent representation of the input
            init_h: Initial embedding of the input
        """
        # Transfer to embedding space
        init_h = self.init_embedding(td)
        bs, num_nodes, emb_dim = init_h.shape
        # (bs*num_nodes, emb_dim)
        update_node_feature = init_h.reshape(-1, emb_dim)
        # shape=(2, num_edges)
        edge_index = self.edge_idx_fn(td, num_nodes)

        for layer in self.gcn_layers[:-1]:
            update_node_feature = layer(update_node_feature, edge_index)
            update_node_feature = F.relu(update_node_feature)
            update_node_feature = F.dropout(
                update_node_feature, training=self.training, p=self.dropout
            )

        # last layer without relu activation and dropout
        update_node_feature = self.gcn_layers[-1](update_node_feature, edge_index)

        # De-batch the graph
        update_node_feature = update_node_feature.view(bs, num_nodes, emb_dim)

        # Residual
        if self.residual:
            update_node_feature = update_node_feature + init_h

        return update_node_feature, init_h



import matplotlib.pyplot as plt
import numpy as np
import torch

from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(td, actions=None, ax=None):
    if ax is None:
        # Create a plot of the nodes
        _, ax = plt.subplots()

    td = td.detach().cpu()

    if actions is None:
        actions = td.get("action", None)

    # If batch_size greater than 0 , we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]
        actions = actions[0]

    locs = td["locs"]

    # Gather locs in order of action if available
    if actions is None:
        log.warning("No action in TensorDict, rendering unsorted locs")
    else:
        actions = actions.detach().cpu()
        locs = gather_by_index(locs, actions, dim=0)

    # Cat the first node to the end to complete the tour
    locs = torch.cat((locs, locs[0:1]))
    x, y = locs[:, 0], locs[:, 1]

    # Plot the visited nodes
    ax.scatter(x, y, color="tab:blue")

    # Add arrows between visited nodes as a quiver plot
    dx, dy = np.diff(x), np.diff(y)
    ax.quiver(x[:-1], y[:-1], dx, dy, scale_units="xy", angles="xy", scale=1, color="k")

    return ax


def render_improvement(td, current_soltuion, best_soltuion):
    coordinates = td["locs"][0]
    real_seq = current_soltuion[:1]
    real_best = best_soltuion[:1]
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 6))  # Create two side-by-side axes

    for ax in [ax1, ax2]:  # Plot on both axes
        if ax == ax1:
            ax.axis([-0.05, 1.05] * 2)
            # plot the nodes
            ax.scatter(
                coordinates[:, 0], coordinates[:, 1], marker="H", s=55, c="blue", zorder=2
            )
            # plot the tour
            real_seq_coordinates = coordinates.gather(
                0, real_seq[0].unsqueeze(1).repeat(1, 2)
            )
            real_seq_coordinates = torch.cat(
                (real_seq_coordinates, real_seq_coordinates[:1]), 0
            )
            ax.plot(
                real_seq_coordinates[:, 0],
                real_seq_coordinates[:, 1],
                color="black",
                zorder=1,
            )
            # mark node
            for i, txt in enumerate(range(real_seq.size(1))):
                ax.annotate(
                    txt,
                    (coordinates[i, 0] + 0.01, coordinates[i, 1] + 0.01),
                )
            ax.set_title("Current Solution")
        else:
            ax.axis([-0.05, 1.05] * 2)
            # plot the nodes
            ax.scatter(
                coordinates[:, 0], coordinates[:, 1], marker="H", s=55, c="blue", zorder=2
            )
            # plot the tour
            real_best_coordinates = coordinates.gather(
                0, real_best[0].unsqueeze(1).repeat(1, 2)
            )
            real_best_coordinates = torch.cat(
                (real_best_coordinates, real_best_coordinates[:1]), 0
            )
            ax.plot(
                real_best_coordinates[:, 0],
                real_best_coordinates[:, 1],
                color="black",
                zorder=1,
            )
            # mark node
            for i, txt in enumerate(range(real_seq.size(1))):
                ax.annotate(
                    txt,
                    (coordinates[i, 0] + 0.01, coordinates[i, 1] + 0.01),
                )
            ax.set_title("Best Solution")
    plt.tight_layout()

    return ax



import abc

from rl4co.models.common.constructive.base import ConstructiveEncoder


class NonAutoregressiveEncoder(ConstructiveEncoder, metaclass=abc.ABCMeta):
    """Template class for an autoregressive encoder, simple wrapper around
    :class:`rl4co.models.common.constructive.base.ConstructiveEncoder`.

    Tip:
        This class will not work as it is and is just a template.
        An example for autoregressive encoder can be found as :class:`rl4co.models.zoo.am.encoder.AttentionModelEncoder`.
    """



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange

from rl4co.models.nn.attention import MultiHeadAttention
from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.ops import Normalization, TransformerFFN


def apply_weights_and_combine(dots, v, tanh_clipping=0):
    # scale to avoid numerical underflow
    logits = dots / dots.std()
    if tanh_clipping > 0:
        # tanh clipping to avoid explosions
        logits = torch.tanh(logits) * tanh_clipping
    # shape: (batch, num_heads, row_cnt, col_cnt)
    weights = nn.Softmax(dim=-1)(logits)
    weights = weights.nan_to_num(0)
    # shape: (batch, num_heads, row_cnt, qkv_dim)
    out = torch.matmul(weights, v)
    # shape: (batch, row_cnt, num_heads, qkv_dim)
    out = rearrange(out, "b h s d -> b s (h d)")
    return out


class MixedScoreFF(nn.Module):
    def __init__(self, num_heads, ms_hidden_dim: int = 32, bias: bool = False) -> None:
        super().__init__()

        self.lin1 = nn.Linear(2 * num_heads, num_heads * ms_hidden_dim, bias=bias)
        self.lin2 = nn.Linear(num_heads * ms_hidden_dim, num_heads, bias=bias)

    def forward(self, dot_product_score, cost_mat_score):
        # dot_product_score shape: (batch, head_num, row_cnt, col_cnt)
        # cost_mat_score shape: (batch, head_num, row_cnt, col_cnt)
        # shape: (batch, head_num, row_cnt, col_cnt, 2)
        two_scores = torch.stack((dot_product_score, cost_mat_score), dim=-1)
        two_scores = rearrange(two_scores, "b h r c s -> b r c (h s)")
        # shape: (batch, row_cnt, col_cnt, 2 * num_heads)
        ms1 = self.lin1(two_scores)
        ms1_activated = F.relu(ms1)
        # shape: (batch, row_cnt, col_cnt, num_heads)
        ms2 = self.lin2(ms1_activated)
        # shape: (batch, row_cnt, head_num, col_cnt)
        mixed_scores = rearrange(ms2, "b r c h -> b h r c")

        return mixed_scores


class EfficientMixedScoreMultiHeadAttention(nn.Module):
    def __init__(self, embed_dim: int, num_heads: int, bias: bool = False):
        super().__init__()

        qkv_dim = embed_dim // num_heads

        self.num_heads = num_heads
        self.qkv_dim = qkv_dim
        self.norm_factor = 1 / math.sqrt(qkv_dim)

        self.Wqv1 = nn.Linear(embed_dim, 2 * embed_dim, bias=bias)
        self.Wkv2 = nn.Linear(embed_dim, 2 * embed_dim, bias=bias)

        # self.init_parameters()
        self.mixed_scores_layer = MixedScoreFF(num_heads, qkv_dim, bias)

        self.out_proj1 = nn.Linear(embed_dim, embed_dim, bias=bias)
        self.out_proj2 = nn.Linear(embed_dim, embed_dim, bias=bias)

    def forward(self, x1, x2, attn_mask=None, cost_mat=None):
        batch_size = x1.size(0)
        row_cnt = x1.size(-2)
        col_cnt = x2.size(-2)

        # Project query, key, value
        q, v1 = rearrange(
            self.Wqv1(x1), "b s (two h d) -> two b h s d", two=2, h=self.num_heads
        ).unbind(dim=0)

        # Project query, key, value
        k, v2 = rearrange(
            self.Wqv1(x2), "b s (two h d) -> two b h s d", two=2, h=self.num_heads
        ).unbind(dim=0)

        # shape: (batch, num_heads, row_cnt, col_cnt)
        dot = self.norm_factor * torch.matmul(q, k.transpose(-2, -1))

        if cost_mat is not None:
            # shape: (batch, num_heads, row_cnt, col_cnt)
            cost_mat_score = cost_mat[:, None, :, :].expand_as(dot)
            dot = self.mixed_scores_layer(dot, cost_mat_score)

        if attn_mask is not None:
            attn_mask = attn_mask.view(batch_size, 1, row_cnt, col_cnt).expand_as(dot)
            dot.masked_fill_(~attn_mask, float("-inf"))

        h1 = self.out_proj1(apply_weights_and_combine(dot, v2))
        h2 = self.out_proj2(apply_weights_and_combine(dot.transpose(-2, -1), v1))

        return h1, h2


class EncoderLayer(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int = 8,
        feedforward_hidden: int = 512,
        normalization: str = "batch",
        bias: bool = False,
    ):
        super().__init__()

        self.op_attn = MultiHeadAttention(embed_dim, num_heads, bias=bias)
        self.ma_attn = MultiHeadAttention(embed_dim, num_heads, bias=bias)
        self.cross_attn = EfficientMixedScoreMultiHeadAttention(
            embed_dim, num_heads, bias=bias
        )

        self.op_ffn = TransformerFFN(embed_dim, feedforward_hidden, normalization)
        self.ma_ffn = TransformerFFN(embed_dim, feedforward_hidden, normalization)

        self.op_norm = Normalization(embed_dim, normalization)
        self.ma_norm = Normalization(embed_dim, normalization)

    def forward(
        self, op_in, ma_in, cost_mat, op_mask=None, ma_mask=None, cross_mask=None
    ):
        op_cross_out, ma_cross_out = self.cross_attn(
            op_in, ma_in, attn_mask=cross_mask, cost_mat=cost_mat
        )
        op_cross_out = self.op_norm(op_cross_out + op_in)
        ma_cross_out = self.ma_norm(ma_cross_out + ma_in)

        # (bs, num_jobs, ops_per_job, d)
        op_self_out = self.op_attn(op_cross_out, attn_mask=op_mask)
        # (bs, num_ma, d)
        ma_self_out = self.ma_attn(ma_cross_out, attn_mask=ma_mask)

        op_out = self.op_ffn(op_cross_out, op_self_out)
        ma_out = self.ma_ffn(ma_cross_out, ma_self_out)

        return op_out, ma_out


class Encoder(nn.Module):
    def __init__(
        self,
        embed_dim: int = 256,
        num_heads: int = 16,
        num_layers: int = 5,
        normalization: str = "batch",
        feedforward_hidden: int = 512,
        init_embedding: nn.Module = None,
        init_embedding_kwargs: dict = {},
        bias: bool = False,
    ):
        super().__init__()
        self.d_model = embed_dim

        if init_embedding is None:
            init_embedding = env_init_embedding(
                "matnet", {"embed_dim": embed_dim, **init_embedding_kwargs}
            )
        self.init_embedding = init_embedding
        self.layers = nn.ModuleList(
            [
                EncoderLayer(
                    embed_dim=embed_dim,
                    num_heads=num_heads,
                    feedforward_hidden=feedforward_hidden,
                    normalization=normalization,
                    bias=bias,
                )
                for _ in range(num_layers)
            ]
        )

    def forward(self, td, attn_mask: torch.Tensor = None):
        # [BS, num_machines, emb], [BS, num_operations, emb]
        ops_embed, ma_embed, edge_feat = self.init_embedding(td)
        try:
            # mask padded ops; shape=(bs, ops)
            ops_attn_mask = ~td["pad_mask"]
        except KeyError:
            ops_attn_mask = None
        # padded ops should also be masked in cross attention; shape=(bs, ops, ma)
        # cross_mask = ops_attn_mask.unsqueeze(-1).expand(-1, -1, ma_embed.size(1))
        for layer in self.layers:
            ops_embed, ma_embed = layer(
                ops_embed,
                ma_embed,
                cost_mat=edge_feat,
                op_mask=ops_attn_mask,  # mask padded operations in attention
                ma_mask=None,  # no padding for machines
                cross_mask=None,
            )
        embedding = (ops_embed, ma_embed)
        return embedding, None



# Base environment
from rl4co.envs.common.base import RL4COEnvBase

# EDA
from rl4co.envs.eda import DPPEnv, MDPPEnv

# Graph
from rl4co.envs.graph import FLPEnv, MCPEnv

# Routing
from rl4co.envs.routing import (
    ATSPEnv,
    CVRPEnv,
    CVRPMVCEnv,
    CVRPTWEnv,
    DenseRewardTSPEnv,
    MDCPDPEnv,
    MTSPEnv,
    MTVRPEnv,
    OPEnv,
    PCTSPEnv,
    PDPEnv,
    PDPRuinRepairEnv,
    SDVRPEnv,
    SHPPEnv,
    SPCTSPEnv,
    SVRPEnv,
    TSPEnv,
    TSPkoptEnv,
)

# Scheduling
from rl4co.envs.scheduling import FFSPEnv, FJSPEnv, JSSPEnv, SMTWTPEnv

# Register environments
ENV_REGISTRY = {
    "atsp": ATSPEnv,
    "cvrp": CVRPEnv,
    "cvrptw": CVRPTWEnv,
    "cvrpmvc": CVRPMVCEnv,
    "dpp": DPPEnv,
    "ffsp": FFSPEnv,
    "jssp": JSSPEnv,
    "fjsp": FJSPEnv,
    "mdpp": MDPPEnv,
    "mtsp": MTSPEnv,
    "op": OPEnv,
    "pctsp": PCTSPEnv,
    "pdp": PDPEnv,
    "pdp_ruin_repair": PDPRuinRepairEnv,
    "sdvrp": SDVRPEnv,
    "shpp": SHPPEnv,
    "svrp": SVRPEnv,
    "spctsp": SPCTSPEnv,
    "tsp": TSPEnv,
    "smtwtp": SMTWTPEnv,
    "mdcpdp": MDCPDPEnv,
    "mtvrp": MTVRPEnv,
    "tsp_kopt": TSPkoptEnv,
    "mcp": MCPEnv,
    "flp": FLPEnv,
}


def get_env(env_name: str, *args, **kwargs) -> RL4COEnvBase:
    """Get environment by name.

    Args:
        env_name: Environment name
        *args: Positional arguments for environment
        **kwargs: Keyword arguments for environment

    Returns:
        Environment
    """
    env_cls = ENV_REGISTRY.get(env_name, None)
    if env_cls is None:
        raise ValueError(
            f"Unknown environment {env_name}. Available environments: {ENV_REGISTRY.keys()}"
        )
    return env_cls(*args, **kwargs)
